# -*- coding: utf-8 -*-
'''
This is a series of custom functions for the inferring of GRN from single cell RNA-seq data.

Codes were written by Kenji Kamimoto.


'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing

import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from scipy import stats

from tqdm.auto import tqdm
from .network_analysis_utility import linkList_to_networkgraph

#import seaborn as sns

settings = {"save_figure_as": "png"}

###################################
### Analyze degree distribution ###
###################################

def _get_degree_distribution(links, cluster):
    g = linkList_to_networkgraph(links.filtered_links[cluster])
    degree_df = _get_degree_info_from_NG(g)
    return degree_df


def plot_degree_distributions(links, plot_model=False, save=None):
    """
    Plot the distribution of network degree (the number of edge per gene).
    Scatter plots in both linear scale and log scale will be generated.

    Args:
        links (Links object): See network_analisis.Links class for detail.
        plot_model (bool): Whether to plot linear approximation line.
        save (str): Folder path to save plots. If the folde does not exist in the path, the function create the folder.
            If None plots will not be saved. Default is None.

    """
    for i in links.cluster:
        print(i)

        g = linkList_to_networkgraph(links.filtered_links[i])
        degree_df = _get_degree_info_from_NG(g)

        if not save is None:
            os.makedirs(save, exist_ok=True)
            path = os.path.join(save, f"degree_dist_{links.name}_{links.threshold_number}_{i}.{settings['save_figure_as']}")
        else:
            path = None

        _plotDegreedist(degree_df=degree_df, plot_model=plot_model, path=path)



def _plotDegreedist(degree_df, plot_model=False, path=None):

    """
    Args:
        degree_df (pandas.DataFrame): data_frame that include degree.
            degree info shold be stored in the column, "degree"

        plot_model (bool): Whether to plot linear approximation line.

        path (str): Folder path to save plots. If the folde does not exist in the path, the function create the folder.
            If None, plots will not be saved. Default is None.
    """

    from sklearn.linear_model import LinearRegression as lr
    df = degree_df.copy()

    dist = df.degree.value_counts()/df.degree.value_counts().sum()
    dist.index = dist.index.astype(int)

    fig, ax = plt.subplots(1,2)

    ax[0].scatter(dist.index.values, dist.values, c="black")
    ax[0].set_title("degree distribution")
    ax[0].set_xlabel("k")
    ax[0].set_ylabel("P(k)")


    #plt.yscale('log')
    #plt.xscale('log')

    x = np.log(dist.index.values).reshape([-1,1])
    y = np.log(dist.values).reshape([-1,1])
    if plot_model:
        model = lr()
        model.fit(x,y)
        x_ = np.array([-1, 5]).reshape([-1,1])
        y_ = model.predict(x_)

        ax[1].set_title(f"degree distribution (log scale)\nslope: {model.coef_[0][0] :.4g}, r2: {model.score(x,y) :.4g}")
        ax[1].plot(x_.flatten(), y_.flatten(), c="black", alpha=0.5)
    else:
        ax[1].set_title(f"degree distribution (log scale)")

    ax[1].scatter(x.flatten(), y.flatten(), c="black")
    ax[1].set_ylim([y.min()-0.2, y.max()+0.2])
    ax[1].set_xlim([-0.2, x.max()+0.2])
    ax[1].set_xlabel("log k")
    ax[1].set_ylabel("log P(k)")

    if path is not None:
        fig.savefig(path, transparent=True)
    plt.show()
    

def _get_degree_info_from_NG(network_x_graph):

    df = pd.DataFrame(np.array(network_x_graph.degree))
    df.columns = ["ind", "degree"]
    df = df.set_index("ind")

    return df

def plot_score_discributions(links, values=None, method="boxplot", save=None):
    """
    Plot the distribution of network scores listed below.
    Indivisual data point is network edge (gene) of GRN in each cluster.

    Network scores
        ['clustering_coefficient', 'clustering_coefficient_weighted',
         'degree_centrality_all', 'degree_centrality_in', 'degree_centrality_out',
         'betweenness_centrality', 'closeness_centrality',
         'eigenvector_centrality', 'assortative_coefficient',
         'average_path_length']

    Args:
        links (Links object): See network_analisis.Links class for detail.
        values (list of str): The list of netwrok score type. If it is None, all network score (listed above) will be used.
        method (str): Plotting method. Select one from "boxplot" or "barplot".
        save (str): Folder path to save plots. If the folde does not exist in the path, the function create the folder.
            If None plots will not be saved. Default is None.
    """
    col = ['clustering_coefficient',
   'clustering_coefficient_weighted', 'degree_centrality_all',
   'degree_centrality_in', 'degree_centrality_out',
   'betweenness_centrality', 'closeness_centrality',
   'eigenvector_centrality', 'assortative_coefficient',
   'average_path_length']
    if values is None:
        values = col
    if method == "boxplot":
        for i in values:
            print(i)
            sns.boxplot(data=links.merged_score,
                        x="cluster", y=i, palette=links.palette.palette.values,
                        order=links.palette.index.values, fliersize=0.0)
            plt.xticks(rotation=90)
            if not save is None:
                os.makedirs(save, exist_ok=True)
                path = os.path.join(save, f"boxplot_{i}_in_{links.name}_{links.threshold_number}.{settings['save_figure_as']}")
                #plt.ylabel("{}\nentropy")
                plt.savefig(path, transparent=True)
            plt.show()
    elif method == "barplot":
        for i in values:
            print(i)
            sns.barplot(data=links.merged_score,
                        x="cluster", y=i, palette=links.palette.palette.values,
                        order=links.palette.index.values)
            plt.xticks(rotation=90)
            if not save is None:
                os.makedirs(save, exist_ok=True)
                path = os.path.join(save, f"barplot_{i}_in_{links.name}_{links.threshold_number}.{settings['save_figure_as']}")
                #plt.ylabel("{}\nentropy")
                plt.savefig(path, transparent=True)
            plt.show()



################################
### Network Entropy analysis ###
################################

def plot_network_entropy_distributions(links, update_network_entropy=False, save=None):
    """
    Plot the distribution of network entropy.
    See the CellOracle paper for the detail.

    Args:
        links (Links object): See network_analisis.Links class for detail.
        values (list of str): The list of netwrok score type. If it is None, all network score (listed above) will be used.
        update_network_entropy (bool): Whether to recalculate network entropy.
        save (str): Folder path to save plots. If the folde does not exist in the path, the function create the folder.
            If None plots will not be saved. Default is None.
    """
    if links.entropy is None:
        links.get_network_entropy()

    if update_network_entropy:
        links.get_network_entropy()

    # fig = plt.figure()

    ax = sns.boxplot(data=links.entropy, x="cluster", y="entropy_norm",
                palette=links.palette.palette.values,
                order=links.palette.index.values, fliersize=0.0)

    ax.tick_params(axis="x", rotation=90)
    ax.set_ylim([0.81,1.0])

    if not save is None:
        os.makedirs(save, exist_ok=True)
        path = os.path.join(save, f"network_entropy_in_{links.name}_{links.threshold_number}.{settings['save_figure_as']}")
        ax.set_ylabel("normalized\nentropy")
        plt.savefig(path, transparent=True)
    plt.show()



# -*- coding: utf-8 -*-


import os, sys

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as colors

import seaborn as sns

from .config import CONFIG


def plot_cluster_whole(self, ax=None, s=CONFIG["s_scatter"], args=CONFIG["default_args"]):

    if ax is None:
        ax = plt

    c = self.colorandum
    ax.scatter(self.embedding[:, 0], self.embedding[:, 1], c=c, s=s, **args)
    ax.axis("off")

def plot_cluster_cells_use(self, ax=None, s=CONFIG["s_scatter"], color=None, show_background=True, args=CONFIG["default_args"]):

    if ax is None:
        ax = plt

    if s == 0:
        color = "white"

    if show_background:
        plot_background(self=self, ax=ax, s=s, args=args)

    if not hasattr(self, "cell_idx_use"):
        self.cell_idx_use = None

    if self.cell_idx_use is None:
        if color is None:
            ax.scatter(self.embedding[:, 0], self.embedding[:, 1], c=self.colorandum, s=s, **args)
        else:
            ax.scatter(self.embedding[:, 0], self.embedding[:, 1], c=color, s=s, **args)

    else:
        if color is None:
            ax.scatter(self.embedding[self.cell_idx_use, 0], self.embedding[self.cell_idx_use, 1],
                       c=self.colorandum[self.cell_idx_use, :],s=s, **args)
        else:
            ax.scatter(self.embedding[self.cell_idx_use, 0], self.embedding[self.cell_idx_use, 1],
                       c=color, s=s, **args)


    ax.axis("off")


def plot_background(self, ax=None, s=CONFIG["s_scatter"], args=CONFIG["default_args"]):

    if ax is None:
        ax = plt

    ax.scatter(self.embedding[:, 0], self.embedding[:, 1], c="lightgray", s=s, **args)

    #ax.set_title("Pseudotime")
    ax.axis("off")

def plot_pseudotime(self, ax=None, s=CONFIG["s_scatter"], show_background=True, cmap="rainbow", args=CONFIG["default_args"]):

    if ax is None:
        ax = plt

    if show_background:
        plot_background(self=self, ax=ax, s=s, args=args)

    if self.cell_idx_use is None:
        ax.scatter(self.embedding[:, 0], self.embedding[:, 1], c=self.pseudotime, cmap=cmap, s=s, **args)
    else:
        ax.scatter(self.embedding[self.cell_idx_use, 0], self.embedding[self.cell_idx_use, 1],
                    c=self.pseudotime[self.cell_idx_use], cmap=cmap, s=s, **args)

    ax.axis("off")


def plot_background_on_grid(self, ax=None, s=CONFIG["s_grid"], args={}):


    if ax is None:
        ax = plt

    if hasattr(self, "mass_filter_whole_reference"):
        mass_filter = self.mass_filter_whole_reference
    elif hasattr(self, "mass_filter_whole"):
        mass_filter = self.mass_filter_whole

    ax.scatter(self.gridpoints_coordinates[:, 0],
               self.gridpoints_coordinates[:, 1], s=0)

    if "c" not in args.keys():
        ax.scatter(self.gridpoints_coordinates[~mass_filter, 0],
               self.gridpoints_coordinates[~mass_filter, 1],
               c="lightgray", s=s, **args)
    else:
        ax.scatter(self.gridpoints_coordinates[~mass_filter, 0],
               self.gridpoints_coordinates[~mass_filter, 1],
               s=s, **args)
    ax.axis("off")


def plot_pseudotime_on_grid(self, ax=None, s=CONFIG["s_grid"], show_background=True, cmap="rainbow", args={}):

    if ax is None:
        ax = plt

    if hasattr(self, "mass_filter_simulation"):
        mass_filter = self.mass_filter_simulation
    elif hasattr(self, "mass_filter"):
        mass_filter = self.mass_filter

    if show_background:
        plot_background_on_grid(self=self, ax=ax, s=s, args=args)
    else:
        plot_cluster_cells_use(self=self, ax=ax, s=0, color="white", show_background=False, args={})


    ax.scatter(self.gridpoints_coordinates[~mass_filter, 0],
               self.gridpoints_coordinates[~mass_filter, 1],
               c=self.pseudotime_on_grid[~mass_filter],
               cmap=cmap, s=s, **args)

    ax.axis("off")

def plot_selected_pseudotime_on_grid(self, ax=None, pseudotime_selected=[], s=CONFIG["s_grid"], show_background=True, args={}):

    if ax is None:
        ax = plt

    mass_filter = self.mass_filter_simulation

    if show_background:
        plot_background_on_grid(self=self, ax=ax, s=s,
                                args={"facecolor": "None",
                                      "c": "None",
                                      "edgecolors":'black',
                                      "linewidths": 0.05})
    else:
        plot_cluster_cells_use(self=self, ax=ax, s=0, color="white", show_background=False, args={})

    x = self.gridpoints_coordinates[~mass_filter, 0]
    y = self.gridpoints_coordinates[~mass_filter, 1]
    for label, color in zip(["True", "False"], ["#EC7063", "#D0D3D4"]):
        if label == "True":
            idx = self.inner_product_df.pseudotime_id.isin(pseudotime_selected).values
        else:
            idx = ~self.inner_product_df.pseudotime_id.isin(pseudotime_selected).values

        ax.scatter(x[idx], y[idx], color=color, label=label, s=s, **args)
        ax.legend()

    ax.axis("off")


def plot_reference_flow_on_grid(self, ax=None, scale=CONFIG["scale_dev"], show_background=True, s=CONFIG["s_scatter"], args=CONFIG["default_args_quiver"]):

    if ax is None:
        ax = plt

    if hasattr(self, "mass_filter_simulation"):
        mass_filter = self.mass_filter_simulation
    elif hasattr(self, "mass_filter"):
        mass_filter = self.mass_filter

    if show_background:
        plot_background(self=self, ax=ax, s=s, args=CONFIG["default_args"])
    else:
        plot_cluster_cells_use(self=self, ax=ax, s=0, color="white", show_background=False, args={})

    ax.quiver(self.gridpoints_coordinates[~mass_filter, 0],
              self.gridpoints_coordinates[~mass_filter, 1],
              self.ref_flow[~mass_filter, 0],
              self.ref_flow[~mass_filter, 1],
              scale=scale, **args)

    ax.axis("off")
def plot_simulation_flow_on_grid(self, ax=None, scale=CONFIG["scale_simulation"], show_background=True, s=CONFIG["s_scatter"], args=CONFIG["default_args_quiver"]):
    _plot_simulation_flow_on_grid(self=self, ax=ax, scale=scale, show_background=show_background, s=s, data_random=False, args=args)

def plot_simulation_flow_random_on_grid(self, ax=None, scale=CONFIG["scale_simulation"], show_background=True, s=CONFIG["s_scatter"], args=CONFIG["default_args_quiver"]):
    _plot_simulation_flow_on_grid(self=self, ax=ax, scale=scale, show_background=show_background, s=s, data_random=True, args=args)

def _plot_simulation_flow_on_grid(self, ax=None, scale=CONFIG["scale_simulation"], show_background=True, s=CONFIG["s_scatter"], data_random=False, args=CONFIG["default_args_quiver"]):

    if ax is None:
        ax = plt

    if show_background:
        plot_background(self=self, ax=ax, s=s, args=CONFIG["default_args"])
    else:
        plot_cluster_cells_use(self=self, ax=ax, s=0, color=None, show_background=False, args={})

    # mass filter selection
    if hasattr(self, "mass_filter_simulation"):
        mass_filter = self.mass_filter_simulation
    elif hasattr(self, "mass_filter"):
        mass_filter = self.mass_filter

    # Gridpoint cordinate selection
    if hasattr(self, "gridpoints_coordinates"):
        gridpoints_coordinates = self.gridpoints_coordinates
    elif hasattr(self, "mass_filter"):
        gridpoints_coordinates = self.flow_grid

    # Arrow selection
    if data_random:
        flow = self.flow_rndm
    else:
        flow = self.flow

    ax.quiver(gridpoints_coordinates[~mass_filter, 0],
              gridpoints_coordinates[~mass_filter, 1],
              flow[~mass_filter, 0],
              flow[~mass_filter, 1], #zorder=20000,
              scale=scale, **args)

    ax.axis("off")

def plot_inner_product_on_grid(self, ax=None, vm=1,s=CONFIG["s_grid"], show_background=True, vmin=None, vmax=None, cmap=None, args={}):

    if ax is None:
        ax = plt

    if vmin is None:
        vmin = -vm
    if vmax is None:
        vmax = vm
    if cmap is None:
        cmap = CONFIG["cmap_ps"]

    try:
        norm = colors.TwoSlopeNorm(vmin=vmin, vcenter=0, vmax=vmax)
    except:
        norm = colors.DivergingNorm(vmin=vmin, vcenter=0, vmax=vmax)

    if show_background:
        plot_background_on_grid(self=self, ax=ax, s=s,
                                args={"facecolor": "None",
                                      "c": "None",
                                      "edgecolors":'black',
                                      "linewidths": 0.05})
    else:
        plot_cluster_cells_use(self=self, ax=ax, s=0, color=None, show_background=False, args={})

    ax.scatter(self.gridpoints_coordinates[~self.mass_filter_simulation, 0],
               self.gridpoints_coordinates[~self.mass_filter_simulation, 1],
               c=self.inner_product[~self.mass_filter_simulation],
               cmap=cmap, norm=norm,#vmin=vmin, vmax=vmax,
               s=s, **args)

    ax.axis("off")

def plot_inner_product_random_on_grid(self, ax=None, vm=1,s=CONFIG["s_grid"], show_background=True, vmin=None, vmax=None, cmap=None, args={}):

    if ax is None:
        ax = plt

    if vmin is None:
        vmin = -vm
    if vmax is None:
        vmax = vm
    if cmap is None:
        cmap = CONFIG["cmap_ps"]

    try:
        norm = colors.TwoSlopeNorm(vmin=vmin, vcenter=0, vmax=vmax)
    except:
        norm = colors.DivergingNorm(vmin=vmin, vcenter=0, vmax=vmax)

    if show_background:
        plot_background_on_grid(self=self, ax=ax, s=s,
                                args={"facecolor": "None",
                                      "c": "None",
                                      "edgecolors":'black',
                                      "linewidths": 0.05})
    else:
        plot_cluster_cells_use(self=self, ax=ax, s=0, color=None, show_background=False, args={})

    ax.scatter(self.gridpoints_coordinates[~self.mass_filter_simulation, 0],
               self.gridpoints_coordinates[~self.mass_filter_simulation, 1],
               c=self.inner_product_random[~self.mass_filter_simulation],
               cmap=cmap, norm=norm,#vmin=vmin, vmax=vmax,
               s=s, **args)

    ax.axis("off")


def plot_inner_product_on_pseudotime(self, ax=None, vm=1, s=CONFIG["s_grid"], vmin=None, vmax=None, cmap=None, args={}):

    if ax is None:
        fig, ax = plt.subplots()

    if vmin is None:
        vmin = -vm
    if vmax is None:
        vmax = vm
    if cmap is None:
        cmap = CONFIG["cmap_ps"]

    try:
        norm = colors.TwoSlopeNorm(vmin=vmin, vcenter=0, vmax=vmax)
    except:
        norm = colors.DivergingNorm(vmin=vmin, vcenter=0, vmax=vmax)

    pcm = ax.scatter(self.pseudotime_on_grid[~self.mass_filter_simulation],
                     self.inner_product[~self.mass_filter_simulation],
                     c=self.inner_product[~self.mass_filter_simulation],
                     cmap=cmap,
                     norm=norm,#vmin=vmin, vmax=vmax,
                     s=s, **args)

    ax.set_ylim([vmin*1.1, vmax*1.1])
    ax.axhline(0, color="lightgray")
    pp = plt.colorbar(pcm, ax=ax, orientation="vertical")
    sns.despine()
    ax.set_xlabel("pseudotime")
    ax.set_ylabel("inner product score")

def plot_inner_product_as_box(self, ax=None, vm=1, vmin=None, vmax=None, args={}):

    if ax is None:
        fig, ax = plt.subplots()

    if vmin is None:
        vmin = -vm
    if vmax is None:
        vmax = vm

    sns.boxplot(data=self.inner_product_df, x="pseudotime_id", y="score", color="white", ax=ax)
    ax.set_xlabel("Digitized_pseudotime")
    ax.set_ylabel("inner product score")
    ax.axhline(0, color="gray")
    ax.set_ylim([vmin*1.1, vmax*1.1])
    ax.tick_params(
                labelleft=False)
    sns.despine()

def plot_quiver(self, ax=None, scale=CONFIG["scale_simulation"], color=None, s=CONFIG["s_scatter"], show_background=True, args=CONFIG["default_args"]):
    _plot_quiver(self=self, ax=ax, scale=scale, color=color, s=s, show_background=show_background, args=args, data_random=False)

def plot_quiver_random(self, ax=None, scale=CONFIG["scale_simulation"], color=None, s=CONFIG["s_scatter"], show_background=True, args=CONFIG["default_args"]):
    _plot_quiver(self=self, ax=ax, scale=scale, color=color, s=s, show_background=show_background, args=args, data_random=True)

def _plot_quiver(self, ax=None, scale=CONFIG["scale_simulation"], color=None, s=CONFIG["s_scatter"], show_background=True, args=CONFIG["default_args"], data_random=False):

    if ax is None:
        ax = plt

    if not hasattr(self, "cell_idx_use"):
        self.cell_idx_use = None

    if self.cell_idx_use is None:
        ix_choice = np.arange(self.embedding.shape[0])
    else:
        ix_choice = self.cell_idx_use

    # Plot whole cell with lightgray
    if show_background:
        ax.scatter(self.embedding[:, 0], self.embedding[:, 1],
                   c="lightgray", alpha=1, s=s, **args)


    ax.scatter(self.embedding[ix_choice, 0], self.embedding[ix_choice, 1],
               c="lightgray", alpha=0.2, edgecolor=(0,0,0,1), s=s, **args)



    if color is None:
        color=self.colorandum[ix_choice]

    quiver_kwargs=dict(headaxislength=7, headlength=11, headwidth=8,
                       linewidths=0.25, width=0.0045,edgecolors="k",
                       color=color, alpha=1)

    if data_random:
        quiver = self.delta_embedding_random
    else:
        quiver = self.delta_embedding

    ax.quiver(self.embedding[ix_choice, 0], self.embedding[ix_choice, 1],
              quiver[ix_choice, 0],
              quiver[ix_choice, 1],
              scale=scale, **quiver_kwargs)

    ax.axis("off")




def visualize_development_module_layout_2(self, scale_for_pseudotime=CONFIG["scale_dev"],
    scale_for_simulation=CONFIG["scale_simulation"], s=CONFIG["s_scatter"], s_grid=CONFIG["s_grid"], vm=1, show_background=True):

    if self.name is None:
        name = "Selected lineage"
    else:
        name = self.name

    fig, ax = plt.subplots(3, 4, figsize=[20,15])

    ax_ = ax[0, 0]
    plot_cluster_whole(self, ax=ax_, s=s)
    ax_.set_title("Whole population")


    ##
    ax_ = ax[0, 1]
    plot_cluster_cells_use(self, ax=ax_, s=s, color="#EC7063", show_background=show_background)
    ax_.set_title(f"{name}")

    ####
    ax_ = ax[0, 2]
    plot_pseudotime(self, ax=ax_, s=s, show_background=show_background)
    ax_.set_title("Pseudotime")

    ###
    ax_ = ax[0, 3]
    plot_reference_flow_on_grid(self, ax=ax_, scale=scale_for_pseudotime, show_background=show_background, s=s)
    ax_.set_title("Development flow")

    ####
    ax_ = ax[1, 0]
    plot_quiver(self, ax=ax_, scale=scale_for_simulation, color="#EC7063", s=s, show_background=show_background)
    ax_.set_title(f"Perturb simulation \n color: {name}")

    ####
    ax_ = ax[1, 1]
    plot_quiver(self, ax=ax_, scale=scale_for_simulation, color=None, s=s, show_background=show_background)
    ax_.set_title("Perturb simulation \n color: cluster")


    ax_ = ax[1, 2]
    plot_simulation_flow_on_grid(self, ax=ax_, scale=scale_for_simulation, show_background=show_background, s=s)
    ax_.set_title("Perturb simulation")


    #####
    ax_ = ax[1, 3]
    plot_cluster_cells_use(self, ax=ax_, s=s, color="#EC7063", show_background=show_background)
    plot_simulation_flow_on_grid(self, ax=ax_, scale=scale_for_simulation, show_background=False, s=s)
    ax_.set_title("Perturb simulation")


    ax_ = ax[2, 0]
    plot_inner_product_on_grid(self, ax=ax_, vm=vm,s=s_grid, show_background=show_background)
    ax_.set_title("Inner product of \n Perturb simulation * Development flow")

    ax_ = ax[2, 1]
    plot_inner_product_on_grid(self, ax=ax_, vm=vm,s=s_grid, show_background=show_background)
    plot_simulation_flow_on_grid(self, ax=ax_, scale=scale_for_simulation, show_background=False, s=s)
    ax_.set_title("Inner product of \n Perturb simulation * Development flow")


    ax_ = ax[2, 2]
    plot_inner_product_on_pseudotime(self, ax=ax_, vm=vm, s=s_grid)

    ax_ = ax[2, 3]
    plot_inner_product_as_box(self, ax=ax_, vm=vm)

    return fig


def visualize_development_module_layout_1(self, scale_for_pseudotime=CONFIG["scale_dev"],
    scale_for_simulation=CONFIG["scale_simulation"], s=CONFIG["s_scatter"], s_grid=CONFIG["s_grid"], vm=1, show_background=True):


    fig, ax = plt.subplots(2, 4, figsize=[20,10])

    ax_ = ax[0, 0]
    plot_cluster_cells_use(self, ax=ax_, s=s, show_background=show_background)

    ##
    ax_ = ax[0, 1]
    plot_pseudotime(self, ax=ax_, s=s, show_background=show_background)

    ####
    ax_ = ax[0, 2]
    plot_pseudotime_on_grid(self, ax=ax_, s=s_grid, show_background=show_background)
    ax_.set_title("Pseudotime on grid")

    ###
    ax_ = ax[0, 3]
    plot_reference_flow_on_grid(self, ax=ax_, scale=scale_for_pseudotime, show_background=show_background, s=s)
    ax_.set_title("Development flow")


    ####
    ax_ = ax[1, 0]
    plot_simulation_flow_on_grid(self, ax=ax_, scale=scale_for_simulation, show_background=show_background, s=s)
    ax_.set_title("Perturb simulation")

    ####
    ax_ = ax[1, 1]
    plot_inner_product_on_grid(self, ax=ax_, vm=vm,s=s_grid, show_background=show_background)
    ax_.set_title("Inner product of \n Perturb simulation * Development flow")


    ax_ = ax[1, 2]
    plot_inner_product_on_pseudotime(self, ax=ax_, vm=vm, s=s_grid)

    #####
    ax_ = ax[1, 3]
    plot_inner_product_as_box(self, ax=ax_, vm=vm)

    return fig

def visualize_development_module_layout_0(self, scale_for_pseudotime=CONFIG["scale_dev"],
    scale_for_simulation=CONFIG["scale_simulation"], s=CONFIG["s_scatter"], s_grid=CONFIG["s_grid"], vm=1, show_background=True):


    fig, ax = plt.subplots(2, 3, figsize=[20, 13.5])

    ax_ = ax[0, 0]
    plot_cluster_cells_use(self, ax=ax_, s=s, show_background=show_background)
    ax_.set_title("Cluster")

    ax_ = ax[0, 1]
    plot_reference_flow_on_grid(self, ax=ax_, scale=scale_for_pseudotime, show_background=show_background, s=s)
    ax_.set_title("Development flow")

    ##
    ax_ = ax[0, 2]
    plot_simulation_flow_on_grid(self, ax=ax_, scale=scale_for_simulation, show_background=show_background, s=s)
    ax_.set_title("Perturb simulation")

    ####
    ax_ = ax[1, 0]
    plot_inner_product_on_grid(self, ax=ax_, vm=vm,s=s_grid, show_background=show_background)
    ax_.set_title("Inner product of \n Perturb simulation * Development flow")


    ax_ = ax[1, 1]
    plot_inner_product_on_pseudotime(self, ax=ax_, vm=vm, s=s_grid)

    #####
    ax_ = ax[1, 2]
    plot_inner_product_as_box(self, ax=ax_, vm=vm)

    return fig





#####

'''

def plot_legend(labels, palette, ax_):

    for i, label in enumerate(labels):
        ax_.scatter([0], [i], s=100, c=palette[label])
        ax_.text(1, i-len(labels)*0.015, s=label)
    ax_.set_ylim([-1, len(labels)])
    ax_.set_xlim([-1, 10])
    ax_.axis("off")

def plot_stackedvar(df, ax, palette=None):

    bottom_feats=[]
    if palette is None:
        for i, j in enumerate(df.index.values):
            if i==0:
                ax.bar(df.columns.values, df.loc[j].values, edgecolor='white', label=j)
            else:
                ax.bar(df.columns.values, df.loc[j].values, label=j,
                        bottom=df.loc[bottom_feats].sum(axis=0).values,
                        edgecolor='white')
            bottom_feats.append(j)
    else:
        for i, j in enumerate(df.index.values):
            if i==0:
                ax.bar(df.columns.values, df.loc[j].values,
                    edgecolor='white', color=palette[j], label=j)
            else:
                ax.bar(df.columns.values, df.loc[j].values, label=j, color=palette[j],
                        bottom=df.loc[bottom_feats].sum(axis=0).values,
                        edgecolor='white')
            bottom_feats.append(j)
    #plt.legend()
    ax.set_xticks(df.columns)

'''



# -*- coding: utf-8 -*-
'''
This file contains custom functions for the analysis of ATAC-seq data.
Genomic activity information (peak of ATAC-seq) will be extracted first.
Then the peak DNA sequence will be subjected to TF motif scan.
Finally we will get list of TFs that potentially binds to a specific gene.

Codes were written by Kenji Kamimoto.


'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing

import pandas as pd
import numpy as np
#import matplotlib.pyplot as plt
#import seaborn as sns

import sys
import os
import pickle
import glob
import logging
from copy import deepcopy
from datetime import datetime

from tqdm.auto import tqdm


from .process_bed_file import df_to_list_peakstr
from .tss_annotation import get_tss_info

def integrate_tss_peak_with_cicero(tss_peak, cicero_connections):
    """
    Process output of cicero data and returns DNA peak information for motif analysis in celloracle.
    Please see the celloracle tutorial for more information.

    Args:
        tss_peak (pandas.dataframe): dataframe about TSS information. Please use the function, "get_tss_info" to get this dataframe.

        cicero_connections (dataframe): dataframe that stores the results of cicero analysis.

    Returns:
        pandas.dataframe: DNA peak about promoter/enhancer and its annotation about target gene.

    """

    # 1. check tss data format and convert if needed
    tss = tss_peak.copy()
    if np.all([i in tss.columns for i in ["chr", "start", "end"]]):
        tss = pd.DataFrame({"peak_id": df_to_list_peakstr(tss),
                            "gene_short_name": tss.gene_short_name.values})
    elif "peak_id" in tss.columns:
        pass
    else:
        raise ValueError("tss_peak format error")

    # 2. process cicero coaccess data
    cicero_tss = pd.merge(cicero_connections, tss.rename(columns={"peak_id": "Peak1"}), on="Peak1", how="inner")
    cicero_tss = cicero_tss.rename(columns={"Peak2": "peak_id"})[["peak_id", "gene_short_name", "coaccess"]]
    cicero_tss = cicero_tss[cicero_tss.coaccess>0]

    # 3. merge tss data and cicero data
    tss["coaccess"] = 1 # Set coaccecss stcore of peaks in tss as 1 before merging data
    merged = pd.concat([cicero_tss, tss], axis=0)
    merged = merged.groupby(by=["peak_id", "gene_short_name"]).max()
    merged = merged.reset_index()

    return merged



# -*- coding: utf-8 -*-
"""
The :mod:`.visualizations` module has several functions that support celloracle.

"""

#from .interactive_simulation_and_plot import Oracle_extended, DEFAULT_PARAMETERS
#from .development_analysis import Oracle_development_module, subset_oracle_for_development_analysiis


__all__ = []



# -*- coding: utf-8 -*-
"""
The :mod:`.go_analysis` module implements Gene Ontology analysis.
This module use goatools internally.

"""

from .goatools_wrapper import geneSymbol2ID, geneID2Symbol, get_GO


__all__ = ["geneSymbol2ID", "geneID2Symbol", "get_GO"]



# -*- coding: utf-8 -*-
"""
The :mod:`.data_conversion` module implements data conversion between different platform.

"""

from .process_seurat_object import seurat_object_to_anndata


__all__ = ["seurat_object_to_anndata"]



# -*- coding: utf-8 -*-

from pathlib import Path
import os

CELLORACLE_DATA_DIR = os.path.join(str(Path.home()), "celloracle_data")
WEB_PAR_DIR = "https://raw.githubusercontent.com/morris-lab/CellOracle/master/celloracle/data"



# -*- coding: utf-8 -*-


import celloracle as co
import pandas as pd
import igraph

def test_get_network_score():
    """
    Several version of the igraph package has incompatibility with Celloracle.
    The function will check whether the igraph package is working.
    """
    print("igraph version: ", igraph.__version__)
    links = co.data.load_tutorial_links_object()
    links.filter_links(p=0.001, weight="coef_abs", threshold_number=2000)
    links.get_network_score()
    assert isinstance(links.merged_score, pd.core.frame.DataFrame)



# -*- coding: utf-8 -*-


CONFIG = {"default_args": {"lw": 0.3, "rasterized": True},
          "s_scatter": 5,
          "s_grid": 20,
          "scale_simulation": 30,
          "scale_dev": 30,
          "cmap_ps": "PiYG",
          "default_args_quiver": {"linewidths": 0.25, "width": 0.004}}



# -*- coding: utf-8 -*-



import io
import logging
import os
import pickle
import subprocess
import sys

import pandas as pd
import numpy as np
from tqdm.auto import tqdm
import matplotlib.pyplot as plt

#import scanpy as sc
#import seaborn as sns
import h5py

#ATTRS = ["delta_embedding", "delta_embedding_random"]

class Oracle_data_strage():
    def __init__(self):
        pass

    def create_hdf_path(self, path):
        if not path.endswith(".hdf5"):
            raise ValueError("path should ends with '.hdf5'")
        self.path = path
        self.path_df = self.path.replace(".hdf5", "_df.hdf5")

        self.names = []
        with h5py.File(self.path, mode='w') as f:
            pass

    def set_hdf_path(self, path, create_if_not_exist=True):
        if not path.endswith(".hdf5"):
            raise ValueError("path should ends with '.hdf5'")
        self.path = path
        self.path_df = self.path.replace(".hdf5", "_df.hdf5")

        try:
            with h5py.File(self.path, mode='r') as f:
                self.names = []
                f.visit(self.names.append)

            unique_genes = np.unique([i.split("/")[0] for i in self.names])
            #print(f"hdf file with {len(unique_genes)} data was found.")
        except:
            print("No hdf file found in the path. New hdf5 file was created.")
            self.create_hdf_path(path=path)

    def save_data(self, oracle, place, attributes):
        """
        Args:
            place (str): Directry in hdf5 file.
            attributes (list of str): attributes to save.
        """

        with h5py.File(self.path, mode='r+') as f:
            for i, j in enumerate(attributes):
                name_ = f"{place}/{j}"
                if name_ in self.names:
                    del f[name_]
                att = getattr(oracle, j)
                try:
                    f[name_] = att
                except:
                    f[name_] = att.astype(h5py.string_dtype(encoding='utf-8'))
                self.names.append(name_)
        self.names = list(set(self.names))

    def load_data(self, oracle, place, attributes):
        with h5py.File(self.path, mode='r') as f:
            for i, j in enumerate(attributes):
                val = f[f"{place}/{j}"][...]
                setattr(oracle, j, val)

    def save_dfs(self, oracle, place, attributes):
        """
        Args:
            place (str): Directry in hdf5 file.
            attributes (list of str): attributes to save.
        """
        for i, j in enumerate(attributes):
            name_ = f"{place}/{j}"
            getattr(oracle, j).to_hdf(self.path_df, key=name_)

    def load_dfs(self, oracle, place, attributes):
        """
        Args:
            place (str): Directry in hdf5 file.
            attributes (list of str): attributes to load.
        """
        for i, j in enumerate(attributes):
            name_ = f"{place}/{j}"
            setattr(oracle, j, pd.read_hdf(self.path_df, key=name_))



# -*- coding: utf-8 -*-
'''
This file contains custom functions for the analysis of ATAC-seq data.
Genomic activity information (peak of ATAC-seq) will be extracted first.
Then the peak DNA sequence will be subjected to TF motif scan.
Finally we will get list of TFs that potentially binds to a specific gene.

Codes were written by Kenji Kamimoto.


'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing

import pandas as pd
import numpy as np
import scanpy as sc
import sys, os
from ..data import __path__ as parent_path
from ..data.config import CELLORACLE_DATA_DIR, WEB_PAR_DIR
from ..utility import load_hdf5
from ..utility.data_download_from_web import download_data_if_data_not_exist


def load_mouse_scATAC_atlas_base_GRN(version="0.10.0", force_download=False):
    """
    Load Transcription factor binding information made from mouse scATAC-seq atlas dataset.
    mm9 genome was used for the reference genome.

    Args:

    Returns:
        pandas.dataframe: TF binding info.
    """
    if version == "0.9.0":
        filename = "TFinfo_data/mm9_mouse_atac_atlas_data_TSS_and_cicero_0.9_accum_threshold_10.5_DF_peaks_by_TFs.parquet"
    elif version == "0.10.0":
        filename = "TFinfo_data/mm9_mouse_atac_atlas_data_TSS_and_cicero_0.9_accum_threshold_10.5_DF_peaks_by_TFs_v202204.parquet"
    else:
        raise ValueError("This version is not found.")

    # Load data from local directory if file exits.
    path = os.path.join(parent_path[0], filename)
    if (force_download == False) & os.path.isfile(path):
        pass
    else:
        path = os.path.join(CELLORACLE_DATA_DIR, filename)
        backup_url = os.path.join(WEB_PAR_DIR, filename)
        download_data_if_data_not_exist(path=path, backup_url=backup_url)

    return pd.read_parquet(path)

load_TFinfo_df_mm9_mouse_atac_atlas = load_mouse_scATAC_atlas_base_GRN # Old function name

def load_Paul2015_data(version="0.10.0", force_download=False):
    """
    Load mouse hematopoiesis scRNA-seq data. The data was processed according to the standard CellOracle preprocessing method described in the tutorial.
    """
    if version == "0.9.0":
        filename = "anndata/Paul_etal_v201212.h5ad"
    elif version == "0.10.0":
        filename = "anndata/Paul_etal_v202204.h5ad"
    else:
        raise ValueError("This version is not found.")

    # Load data from local directory if file exits.
    path = os.path.join(parent_path[0], filename)
    if (force_download == False) & os.path.isfile(path):
        pass
    else:
        path = os.path.join(CELLORACLE_DATA_DIR, filename)
        backup_url = os.path.join(WEB_PAR_DIR, filename)
        download_data_if_data_not_exist(path=path, backup_url=backup_url)

    return sc.read_h5ad(path)

def load_tutorial_links_object(version="0.10.0", force_download=False):
    """
    """
    if version == "0.9.0":
        filename = "tutorial_data/links_louvain_190829.celloracle.links"
    elif version == "0.10.0":
        filename = "tutorial_data/links_louvain_v20220406.celloracle.links"
    else:
        raise ValueError("This version is not found.")
    # Load data from local directory if file exits.
    path = os.path.join(parent_path[0], filename)
    if (force_download == False) & os.path.isfile(path):
        pass
    else:
        path = os.path.join(CELLORACLE_DATA_DIR, filename)
        backup_url = os.path.join(WEB_PAR_DIR, filename)
        download_data_if_data_not_exist(path=path, backup_url=backup_url)

    return load_hdf5.load_hdf5(path)

def load_tutorial_oracle_object(version="0.10.0", force_download=False):
    if version == "0.9.0":
        filename = "tutorial_data/Paul_etal_v20210704.celloracle.oracle"
    elif version == "0.10.0":
        filename = "tutorial_data/Paul_etal_v20220406.celloracle.oracle"
    else:
        raise ValueError("This version is not found.")

    # Load data from local directory if file exits.
    path = os.path.join(parent_path[0], filename)
    if (force_download == False) & os.path.isfile(path):
        pass
    else:
        path = os.path.join(CELLORACLE_DATA_DIR, filename)
        backup_url = os.path.join(WEB_PAR_DIR, filename)
        download_data_if_data_not_exist(path=path, backup_url=backup_url)
    return load_hdf5.load_hdf5(path)



# -*- coding: utf-8 -*-
'''
This file contains custom functions for the analysis of ATAC-seq data.
Genomic activity information (peak of ATAC-seq) will be extracted first.
Then the peak DNA sequence will be subjected to TF motif scan.
Finally we will get list of TFs that potentially binds to a specific gene.

Codes were written by Kenji Kamimoto.


'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing

import pandas as pd
import numpy as np
#import matplotlib.pyplot as plt
#import seaborn as sns

import sys
import os
import pickle
import glob
import logging
from copy import deepcopy
from datetime import datetime

from tqdm.auto import tqdm

from genomepy import Genome

from gimmemotifs.motif import Motif
from gimmemotifs.motif import default_motifs
from gimmemotifs.scanner import Scanner

from ..utility.hdf5_processing import dump_hdf5, load_hdf5
from ..utility import save_as_pickled_object, load_pickled_object, intersect,\
                      makelog, inverse_dictionary
#
from .motif_analysis_utility import scan_dna_for_motifs, is_genome_installed
from .process_bed_file import read_bed, peak2fasta, remove_zero_seq
from .motif_data import load_motifs
from .reference_genomes import SUPPORTED_REF_GENOME



def load_TFinfo(file_path):
    """
    Load TFinfo object which was saved as hdf5 file.

    Args:
        file_path (str): file path.

    Returns:
        TFinfo: Loaded TFinfo object.

    """
    return load_hdf5(filename=file_path, obj_class=TFinfo)

def load_TFinfo_from_parquets(folder_path):
    """
    Load TFinfo object which was saved with the function; "save_as_parquet".

    Args:
        folder_path (str): folder path

    Returns:
        TFinfo: Loaded TFinfo object.
    """

    folder_path_ = os.path.abspath(os.path.join(folder_path, "TFinfo_comp"))
    tfi = load_pickled_object(os.path.join(folder_path_, "tfi.pickle"))

    compressed_files_path = glob.glob(os.path.join(folder_path_, "*.parquet"))
    compressed_file_name = [i.split("/")[-1] for i in  compressed_files_path]

    for name, path in zip(compressed_file_name, compressed_files_path):
        if name == "scanned_df.parquet":
            tfi.scanned_df = pd.read_parquet(path)
        if name == "peak_df.parquet":
            tfi.peak_df = pd.read_parquet(path)
        if name == "TF_onehot.parquet":
            tfi.TF_onehot = pd.read_parquet(path)
    return tfi

def make_TFinfo_from_scanned_file(path_to_raw_bed, path_to_scanned_result_bed, ref_genome):
    """
    This function is currently an available.

    """


    peak_df = read_bed(path_to_raw_bed)
    peak_df = peak_df.drop(columns=["chrom", "start", "end"],axis=1)
    peak_df = peak_df.rename(columns={"name":"gene_short_name",
                                      "seqname":"peak_id"})

    tfinfo = TFinfo(peak_data_frame=peak_df, ref_genome=ref_genome)


    tfinfo.all_target_gene = tfinfo.peak_df.gene_short_name.unique()
    tfinfo.all_peaks = tfinfo.peak_df.peak_id.unique()

    tfinfo.scanned_df = read_bed(path_to_scanned_result_bed).rename(columns={"name": "motif_id"})


    return tfinfo



class TFinfo():
    """
    This is a custom class for motif analysis in celloracle.
    TFinfo object performs motif scan using the TF motif database in gimmemotifs and several functions of genomepy.
    Analysis results can be exported as a python dictionary or dataframe.
    These files; python dictionary of dataframe of TF binding information, are needed during GRN inference.

    Attributes:
        peak_df (pandas.dataframe): dataframe about DNA peak and target gene data.
        all_target_gene (array of str): target genes.
        ref_genome (str): reference genome name that was used in DNA peak generation.
        scanned_df (dictionary): Results of motif scan. Key is a peak name. Value is a dataframe of motif scan.
        dic_targetgene2TFs (dictionary): Final product of motif scan. Key is a target gene. Value is a list of regulatory candidate genes.
        dic_peak2Targetgene (dictionary): Dictionary. Key is a peak name. Value is a list of the target gene.
        dic_TF2targetgenes (dictionary): Final product of motif scan. Key is a TF. Value is a list of potential target genes of the TF.

    """

    def __init__(self, peak_data_frame, ref_genome):
        """
        Instantiate TFinfo object.

        Args:
           peak_data_frame (pandas.dataframe): dataframe about DNA peak and target gene data.

           ref_genome (str): reference genome name that was used in DNA peak generation.

        """


        self.easy_log = pd.DataFrame()
        self.__addLog("initiation")

        self.peak_df = peak_data_frame.copy()
        # remove redundant information and check column name
        self.peak_df = self.peak_df.groupby(["peak_id","gene_short_name"]).sum()
        self.peak_df = self.peak_df.reset_index(drop=False)

        self.all_target_gene = self.peak_df.gene_short_name.unique()
        self.all_peaks = self.peak_df.peak_id.unique()

        self.ref_genome = ref_genome

        # check ref_genome is supported or not
        if ref_genome in SUPPORTED_REF_GENOME.ref_genome.values:
            self.species = SUPPORTED_REF_GENOME.species[SUPPORTED_REF_GENOME.ref_genome==ref_genome].values[0]
        else:
            #print(f"ref_genome: {ref_genome} is not supported in celloracle. See celloracle.motif_analysis.SUPPORTED_REF_GENOME to get supported ref genome list. If you have a request for a new referencce genome, please post an issue in github issue page.")
            self.species = "not_in_the_default_species_list"

        # check  genome installation
        if not is_genome_installed(ref_genome=ref_genome):
            raise ValueError(f"ref_genome: {ref_genome} is not installed. TFinfo initiation failed.")

        self.scanned_df = None
        self.TF_onehot = None
        self.scanned_filtered = None
        self.index_to_be_used = None
        self.thresholding_comment = []


    def __addLog(self, info):

        new_df = pd.DataFrame({"time": [datetime.now().ctime()], "info": [info]})
        self.easy_log = pd.concat([self.easy_log, new_df], axis=0).reset_index(drop=True)

    def copy(self):
        """
        Deepcoty itself.
        """
        return deepcopy(self)


    def save_as_parquet(self, folder_path=None):
        """
        Save itself. Some attributes are saved as parquet file.

        Args:
            folder_path (str): folder path
        """

        os.makedirs(folder_path, exist_ok=True)

        tmp_self = self.copy()

        tmp_self.peak_df.to_parquet(os.path.join(folder_path, "peak_df.parquet"))
        tmp_self.peak_df = None

        if not self.scanned_df is None:
            tmp_self.scanned_df.to_parquet(os.path.join(folder_path, "scanned_df.parquet"))
            tmp_self.scanned_df = None

        if not self.TF_onehot is None:
            tmp_self.TF_onehot.to_parquet(os.path.join(folder_path, "TF_onehot.parquet"))
            tmp_self.TF_onehot = None

        self.__addLog("save_as_compressed")

        save_as_pickled_object(tmp_self, os.path.join(folder_path, "tfi.pickle"))

        print(f"file saved in: {folder_path}")


    def to_hdf5(self, file_path):
        """
        Save object as hdf5.

        Args:
            file_path (str): file path to save file. Filename needs to end with '.celloracle.tfinfo'
        """
        if file_path.endswith(".celloracle.tfinfo"):
            pass
        else:
            raise ValueError("Filename needs to end with '.celloracle.tfinfo'")

        compression_opts = 7
        dump_hdf5(obj=self, filename=file_path,
                  data_compression=compression_opts,  chunks=(2048, 2048),
                  noarray_compression=compression_opts, pickle_protocol=4)



    def scan(self, background_length=200, fpr=0.02, n_cpus=-1, verbose=True, motifs=None, TF_evidence_level="direct_and_indirect", TF_formatting="auto", divide=100000):
        """
        Scan DNA sequences searching for TF binding motifs.

        Args:
           background_length (int): background length. This is used for the calculation of the binding score.

           fpr (float): False positive rate for motif identification.

           n_cpus (int): number of CPUs for parallel calculation.

           verbose (bool): Whether to show a progress bar.

           motifs (list): a list of gimmemotifs motifs, will revert to default_motifs() if None

           TF_evidence_level (str): Please select one from ["direct", "direct_and_indirect"]. If "direct" is selected, TFs that have a binding evidence were used.
               If "direct_and_indirect" is selected, TFs with binding evidence and inferred TFs are used.
               For more information, please read explanation of Motif class in gimmemotifs documentation (https://gimmemotifs.readthedocs.io/en/master/index.html)

        """

        self.fpr = fpr
        self.background_length = background_length

        ## 1. initialilze scanner  ##
        # load motif
        if motifs is None:
            if verbose:
                print("No motif data entered. Loading default motifs for your species ...")

            if self.species in ["Mouse", "Human", "Rat"]: # If species is vertebrate, we use gimmemotif default motifs as a default.
                motifs = default_motifs()
                self.motif_db_name = "gimme.vertebrate.v5.0"
                self.TF_formatting = True
                if verbose:
                    print(" Default motif for vertebrate: gimme.vertebrate.v5.0. \n For more information, please see https://gimmemotifs.readthedocs.io/en/master/overview.html \n")

            elif self.species in ["Zebrafish"]: # If species is Zebrafish, we use CisBP database.
                self.motif_db_name = 'CisBP_ver2_Danio_rerio.pfm'
                motifs = load_motifs(self.motif_db_name)
                self.TF_formatting = False
                if verbose:
                    print(f" Default motif for {self.species}: {self.motif_db_name}. \n For more information about the motif data, please see http://cisbp.ccbr.utoronto.ca. \n")

            elif self.species in ["S.cerevisiae"]:
                self.motif_db_name = 'CisBP_ver2_Saccharomyces_cerevisiae.pfm'
                motifs = load_motifs(self.motif_db_name)
                self.TF_formatting = False
                if verbose:
                    print(f" Default motif for {self.species}: {self.motif_db_name}. \n For more information about the motif data, please see http://cisbp.ccbr.utoronto.ca. \n")

            elif self.species in ["Xenopus tropicalis"]:
                #self.motif_db_name = 'CisBP_ver2_Xenopus_tropicalis_and_Xenopus_laevis.pfm' # V0.10.13 or earlier.
                self.motif_db_name = 'CisBP_ver2_Xenopus_tropicalis.pfm' # V0.10.14 or later.
                motifs = load_motifs(self.motif_db_name)
                self.TF_formatting = False
                if verbose:
                    print(f" Default motif for {self.species}: {self.motif_db_name}.")
                    print(f" Default motif for {self.species} was changed at celloracle 0.10.14. \n For more information, please see celloracle documentation Changelog page. \n")

            elif self.species in ["Xenopus laevis"]:
                self.motif_db_name = 'CisBP_ver2_Xenopus_laevis.pfm'
                motifs = load_motifs(self.motif_db_name)
                self.TF_formatting = False
                if verbose:
                    print(f" Default motif for {self.species}: {self.motif_db_name}. \n For more information about the motif data, please see http://cisbp.ccbr.utoronto.ca. \n")

            elif self.species in ["Drosophila"]:
                self.motif_db_name = 'CisBP_ver2_Drosophila_melanogaster.pfm'
                motifs = load_motifs(self.motif_db_name)
                self.TF_formatting = False
                if verbose:
                    print(f" Default motif for {self.species}: {self.motif_db_name}. \n For more information about the motif data, please see http://cisbp.ccbr.utoronto.ca. \n")

            elif self.species in ["C.elegans"]:
                self.motif_db_name = 'CisBP_ver2_Caenorhabditis_elegans.pfm'
                motifs = load_motifs(self.motif_db_name)
                self.TF_formatting = False
                if verbose:
                    print(f" Default motif for {self.species}: {self.motif_db_name}. \n For more information about the motif data, please see http://cisbp.ccbr.utoronto.ca. \n")

            elif self.species in ["Arabidopsis"]:
                self.motif_db_name = 'CisBP_ver2_Arabidopsis_thaliana_GENE_ID.pfm'
                motifs = load_motifs(self.motif_db_name)
                self.TF_formatting = False
                if verbose:
                    print(f" Default motif for {self.species}: {self.motif_db_name}. \n For more information about the motif data, please see http://cisbp.ccbr.utoronto.ca. \n")

            elif self.species in ["Chicken"]:
                self.motif_db_name = 'CisBP_ver2_Gallus_gallus.pfm'
                motifs = load_motifs(self.motif_db_name)
                self.TF_formatting = False
                if verbose:
                    print(f" Default motif for {self.species}: {self.motif_db_name}. \n For more information about the motif data, please see http://cisbp.ccbr.utoronto.ca. \n")

            elif self.species in ["Guinea_Pig"]:
                self.motif_db_name = 'CisBP_ver2_Cavia_porcellus.pfm'
                motifs = load_motifs(self.motif_db_name)
                self.TF_formatting = False
                if verbose:
                    print(f" Default motif for {self.species}: {self.motif_db_name}. \n For more information about the motif data, please see http://cisbp.ccbr.utoronto.ca. \n")

            elif self.species in ["Pig"]:
                self.motif_db_name = 'CisBP_ver2_Sus_scrofa.pfm'
                motifs = load_motifs(self.motif_db_name)
                self.TF_formatting = False
                if verbose:
                    print(f" Default motif for {self.species}: {self.motif_db_name}. \n For more information about the motif data, please see http://cisbp.ccbr.utoronto.ca. \n")

            else:
                raise ValueError(f"We don't have default motifs for your species, Please specify motif data by yourself.")

        else:
            # Check format
            if isinstance(motifs, list):
                if isinstance(motifs[0], Motif):
                    if verbose:
                        print("Checking your motifs... Motifs format looks good. \n")
                else:
                    raise ValueError(f"Motif data type was invalid.")
            else:
                raise ValueError(f"motifs should be a list of Motif object in gimmemotifs.")

            self.motif_db_name = "custom_motifs"
            if TF_formatting == "auto":
                self.TF_formatting = False
            else:
                self.TF_formatting = TF_formatting

        self.motifs = motifs

        self.dic_motif2TFs = _get_dic_motif2TFs(species=self.species, motifs=motifs, TF_evidence_level=TF_evidence_level, formatting=self.TF_formatting)
        self.TF_evidence_level = TF_evidence_level

        # initialize scanner
        if verbose:
            print("Initiating scanner... \n")
        s = Scanner(ncpus=n_cpus)

        # set parameters
        s.set_motifs(motifs)
        try:
            s.set_background(genome=self.ref_genome, size=background_length) # For gimmemotifs ver 14.4
        except:
            s.set_background(genome=self.ref_genome, length=background_length)# For old gimmemotifs ver 13

        #s.set_background(genome="mm9", length=400)
        if verbose:
            print("Calculating FPR-based threshold. This step may take substantial time when you load a new ref-genome. It will be done quicker on the second time. \n")
        s.set_threshold(fpr=fpr)

        ## 2. motif scan ##


        # Get DNA sequences
        target_sequences = peak2fasta(self.all_peaks, self.ref_genome)
        # Remove DNA sequence with zero length
        target_sequences = remove_zero_seq(fasta_object=target_sequences)

        if verbose:
            print("Motif scan started .. It may take long time.\n")
        self.scanned_df = scan_dna_for_motifs(scanner_object=s,
                                              motifs_object=motifs,
                                              sequence_object=target_sequences,
                                              divide=divide,
                                              verbose=verbose)

        self.__addLog("scanMotifs")

    def reset_dictionary_and_df(self):
        """
        Reset TF dictionary and TF dataframe.
        The following attributes will be erased: TF_onehot, dic_targetgene2TFs, dic_peak2Targetgene, dic_TF2targetgenes.

        """
        self.TF_onehot = None
        self.dic_targetgene2TFs = None
        self.dic_peak2Targetgene = None
        self.dic_TF2targetgenes = None

    def reset_filtering(self):
        """
        Reset filtering information.
        You can use this function to stat over the filtering step with new conditions.
        The following attributes will be erased: TF_onehot, dic_targetgene2TFs, dic_peak2Targetgene, dic_TF2targetgenes.

        """
        self.scanned_filtered = None
        self.reset_dictionary_and_df()

    def filter_peaks(self, peaks_to_be_remained):
        """
        Filter peaks.

        Args:
            peaks_to_be_remained (array of str): list of peaks. Peaks that are NOT in the list will be removed.

        """
        if self.scanned_df is None:
            raise ValueError("Motif Scan is not done.")

        if self.scanned_filtered is None:
            self.scanned_filtered = self.scanned_df[["seqname", "motif_id", "score"]].copy()

        before = len(self.scanned_filtered)
        self.scanned_filtered = self.scanned_filtered[self.scanned_filtered.seqname.isin(peaks_to_be_remained)]
        after = len(self.scanned_filtered)
        print(f"peaks were filtered: {before} -> {after}")

        self.reset_dictionary_and_df()

        self.thresholding_comment.append(f"threshold peaks")
        self.__addLog("setThresholding_byBindScore")

    def filter_motifs_by_score(self, threshold, method="cumulative_score"):
        """
        Remove motifs with low binding scores.

        Args:
            method (str): thresholding method. Select either of ["indivisual_score", "cumulative_score"]
        """
        if method == "cumulative_score":
            self._thresholding_by_cumulative_bind_score(threshold_score=threshold)

        elif method == "indivisual_score":
            self._thresholding_by_bind_score(threshold_score=threshold)

        else:
            raise ValueError("Method is wrong. Select from ['indivisual_score', 'cumulative_score'] ")


    def _thresholding_by_bind_score(self, threshold_score):
        if self.scanned_df is None:
            raise ValueError("Motif Scan is not done.")

        if self.scanned_filtered is None:
            self.scanned_filtered = self.scanned_df[["seqname", "motif_id", "score"]].copy()

        before = len(self.scanned_filtered)
        self.scanned_filtered = self.scanned_filtered[self.scanned_filtered.score>=threshold_score]
        after = len(self.scanned_filtered)
        print(f"Filtering finished: {before} -> {after}")

        self.reset_dictionary_and_df()

        self.thresholding_comment.append(f"score_threshold:_{threshold_score}")
        self.__addLog("thresholdingByBindScore")


    def _thresholding_by_cumulative_bind_score(self, threshold_score):
        if self.scanned_df is None:
            raise ValueError("Motif Scan is not done.")

        if self.scanned_filtered is None:
            self.scanned_filtered = self.scanned_df[["seqname", "motif_id", "score"]].copy()

        before = len(self.scanned_filtered)

        tmp = self.scanned_filtered.groupby(by=["seqname", "motif_id"]).sum()
        tmp = tmp[tmp.score >= threshold_score]
        tmp = tmp.reset_index()
        self.scanned_filtered = tmp

        self.reset_dictionary_and_df()

        after = len(self.scanned_filtered)
        print(f"Filtering finished: {before} -> {after}")
        self.__addLog("thresholdingByAcumulatedBindScore")

    def make_TFinfo_dataframe_and_dictionary(self, verbose=True):
        """
        This is the final step of motif_analysis.
        Convert scanned results into a data frame and dictionaries.

        Args:
            verbose (bool): Whether to show a progress bar.
        """
        if self.scanned_filtered is None:
            self.scanned_filtered = self.scanned_df[["seqname", "motif_id", "score"]].copy()

        if verbose:
            print("1. Converting scanned results into one-hot encoded dataframe.")
        self._make_TFinfo_dataframe(verbose=verbose)

        if verbose:
            print("2. Converting results into dictionaries.")
            self._make_dictionaries(verbose=verbose)

    def _make_TFinfo_dataframe(self, verbose=True):
        """
        Convert scanned results into one-hot encoded dataframe.
        rows are peak index, columns are TFs.
        The value is 1 If TF have binding motif in the peak.
        The result will stored in self.TF_onehot

        Args:
            verbose (bool): Whether to show progressbar.

        """

        def _motifs_to_TFs_as_onehot_series(motifs):

            tfs = []
            for motif in motifs:
                tfs += self.dic_motif2TFs[motif]

            tfs = np.unique(tfs)

            series = pd.Series(np.repeat(1, len(tfs)), index=tfs)
            return series


        # preprocess
        peak_list = self.scanned_filtered.seqname.unique()
        multi_index_df = self.scanned_filtered.set_index(["seqname", "motif_id"])

        # process data
        li = []
        if verbose:
            loop = tqdm(peak_list)
        else:
            loop = peak_list
        for peak in loop:
            motifs = multi_index_df.loc[peak].index.values
            series = _motifs_to_TFs_as_onehot_series(motifs)
            li.append(series)

        # get hot encoded Data frame
        self.TF_onehot = pd.concat(li, axis=1, sort=True).transpose()
        del li

        self.TF_onehot = self.TF_onehot.fillna(0).astype("int")
        self.TF_onehot.index = peak_list

        self.__addLog("makeOnehotDataframe")

    def _make_dictionaries(self, verbose=True):
        """
        Convert scanned results into dictionaries.

        The result will stored in self.dic_targetgene2TFs and self.dic_TF2targetgenes

        self.dic_TF2targetgene

        Args:
            verbose (bool): Whether to show progressbar.

        """
        if self.TF_onehot is None:
            raise ValueError("Process has not complete yet.")

        dic_targetgene2TFs ={}
        dic_peak2Targetgene = {}

        if verbose:
            loop = tqdm(self.all_target_gene)
        else:
            loop = self.all_target_gene
        for tg in loop:
            peaks = self.peak_df[self.peak_df.gene_short_name==tg].peak_id.values
            peaks = np.array(intersect(peaks, self.TF_onehot.index))
            #print(peaks)

            tmp_series = self.TF_onehot.loc[peaks].max(axis=0)
            tfs = tmp_series[tmp_series==1].index.values
            dic_targetgene2TFs[tg] = tfs

            for peak in peaks:
                dic_peak2Targetgene[peak] = tg



        self.dic_targetgene2TFs = dic_targetgene2TFs
        self.dic_peak2Targetgene = dic_peak2Targetgene

        self.dic_TF2targetgenes = inverse_dictionary(dictionary=dic_targetgene2TFs,
                                                     verbose=verbose,
                                                     return_value_as_numpy=True)

        self.__addLog("make_dictionaries")


    def to_dataframe(self, verbose=True):
        """
        Return results as a dataframe.
        Rows are peak_id, and columns are TFs.

        Args:
            verbose (bool): Whether to show a progress bar.

        Returns:
            pandas.dataframe: TFinfo matrix.
        """
        if self.TF_onehot is None:
            if verbose:
                print("Converting scanned results into one-hot encoded dataframe ...")
            self._make_TFinfo_dataframe(verbose=verbose)

        tmp_peak_df = self.TF_onehot.reindex(self.peak_df.peak_id.values).fillna(0)
        tmp_peak_df = tmp_peak_df.reset_index(drop=True)

        tmp = pd.concat([self.peak_df, tmp_peak_df], axis=1)


        return tmp



    def to_dictionary(self, dictionary_type="targetgene2TFs", verbose=True):
        """
        Return TF information as a python dictionary.

        Args:
            dictionary_type (str): Type of dictionary. Select from ["targetgene2TFs", "TF2targetgenes"].
                If you chose "targetgene2TFs", it returns a dictionary in which a key is a target gene, and a value is a list of regulatory candidate genes (TFs) of the target.
                If you chose "TF2targetgenes", it returns a dictionary in which a key is a TF and a value is a list of potential target genes of the TF.
        Returns:
            dictionary: dictionary.
        """
        if self.TF_onehot is None:
            self.make_TFinfo_dataframe_and_dictionary(verbose=verbose)

        elif (self.dic_targetgene2TFs is None) | (self.dic_TF2targetgenes is None):
            if verbose:
                print("Converting results into dictionaries.")
            self._make_dictionaries(verbose=verbose)

        if dictionary_type == "targetgene2TFs":
            return self.dic_targetgene2TFs

        elif dictionary_type == "TF2targetgenes":
            return self.dic_TF2targetgenes

        else:
            raise ValueError(f'{dictionary_type} is not available for dictionary_type.\nSelect from ["targetgene2TFs", "TF2targetgenes"]')




######################################################
### 4.2. Make TFinfo dataFrame for GRN inference  ###
######################################################

from gimmemotifs.config import DIRECT_NAME, INDIRECT_NAME


def _get_dic_motif2TFs(species, motifs, TF_evidence_level="direct_and_indirect", formatting=True):
    """

    Args:

    TF_evidence_level (str): Please select one from ["direct", "direct_and_indirect"]. If "direct" is selected, TFs that have a binding evidence were used.
        If "direct_and_indirect" is selected, TFs with binding evidence and inferred TFs are used.
        For more information, please read explanation of Motif class in gimmemotifs documentation (https://gimmemotifs.readthedocs.io/en/master/index.html)

    """

    if TF_evidence_level == "direct_and_indirect":
        factor_kind = [DIRECT_NAME, INDIRECT_NAME]
    elif TF_evidence_level == "direct":
        factor_kind = [DIRECT_NAME]

    dic_motif2TFs = {}

    for i in motifs:
        fcs = []
        for j in factor_kind:
            fcs += i.factors[j]
        dic_motif2TFs[i.id] = fcs

    if formatting:
        if species in ["Mouse", "Rat"]:
            for key in dic_motif2TFs.keys():
                dic_motif2TFs[key] = [tf.capitalize() for tf in dic_motif2TFs[key]]

        elif species in ["Human", "S.cerevisiae", "Arabidopsis"]:
            for key in dic_motif2TFs.keys():
                dic_motif2TFs[key] = [tf.upper() for tf in dic_motif2TFs[key]]

        elif species in ["Zebrafish", "Xenopus"]:
            for key in dic_motif2TFs.keys():
                dic_motif2TFs[key] = [tf.lower() for tf in dic_motif2TFs[key]]

        elif species in ["Drosophila", "C.elegans"]:
            pass

    return dic_motif2TFs



#########################################################
### 4.2. Calculate P-values; Multi processing version ###
#########################################################



#########################
### 5.1 Visualization ###
#########################




######################
### 6. GO analysis ###
######################


#############################
### Scoring with ML model ###
#############################



# -*- coding: utf-8 -*-
"""
The :mod:`.oracle_utility` module has several functions that support celloracle.

"""

from .interactive_simulation_and_plot import Oracle_extended, DEFAULT_PARAMETERS
from .development_analysis import Oracle_development_module, subset_oracle_for_development_analysiis
from .utility import Oracle_data_strage

__all__ = [
           "Oracle_extended",
           "Oracle_data_strage",
           "DEFAULT_PARAMETERS",
           "Oracle_development_module",
           "subset_oracle_for_development_analysiis"
           ]



# -*- coding: utf-8 -*-
'''
This is a series of custom functions for the inferring of GRN from single cell RNA-seq data.

Codes were written by Kenji Kamimoto.


'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing

import os

import numpy as np
import pandas as pd

from scipy import stats

from tqdm.auto import tqdm

from ..network.net_core import Net
from ..utility import standard
from .links_object import Links
from ..trajectory.oracle_utility import _adata_to_df, _get_clustercolor_from_anndata, _check_color_information_and_create_if_not_found


RIDGE_SOLVER = "auto"


########################################
### Construct cluster specific GRNs  ###
########################################

def get_links(oracle_object, cluster_name_for_GRN_unit=None, alpha=10, bagging_number=20, verbose_level=1, test_mode=False, model_method="bagging_ridge", n_jobs=-1):
    """
    Make GRN for each cluster and returns results as a Links object.
    Several preprocessing should be done before using this function.

    Args:
        oracle_object (Oracle): See Oracle module for detail.

        cluster_name_for_GRN_unit (str): Cluster name for GRN calculation. The cluster information should be stored in Oracle.adata.obs.

        alpha (float or int): The strength of regularization.
            If you set a lower value, the sensitivity increases, and you can detect weaker network connections. However, there may be more noise.
            If you select a higher value, it will reduce the chance of overfitting.

        bagging_number (int): The number used in bagging calculation.


        verbose_level (int): if [verbose_level>1], most detailed progress information will be shown.
            if [1 >= verbose_level > 0], one progress bar will be shown.
            if [verbose_level == 0], no progress bar will be shown.

        test_mode (bool): If test_mode is True, GRN calculation will be done for only one cluster rather than all clusters.

        model_method (str): Chose modeling algorithm. "bagging_ridge" or "bayesian_ridge"

        n_jobs (int): Number of cpu cores for parallel calculation.  -1 means using all available cores. Default is -1.

    """
    if model_method not in ["bagging_ridge", "bayesian_ridge"]:
        raise ValueError("model_mothod error. Please set 'bagging_ridge' or 'bayesian_ridge'.")

    if cluster_name_for_GRN_unit is None:
        cluster_name_for_GRN_unit = oracle_object.cluster_column_name

    # calculate GRN for each cluster
    linkLists = _fit_GRN_for_network_analysis(oracle_object, cluster_name_for_GRN_unit=cluster_name_for_GRN_unit,
                                  alpha=alpha, bagging_number=bagging_number,  verbose_level=verbose_level, test_mode=test_mode,
                                  model_method=model_method, n_jobs=n_jobs)

    # initiate links object
    links = Links(name=cluster_name_for_GRN_unit,
                 links_dict=linkLists)

    # extract color infomation
    # update color information
    _check_color_information_and_create_if_not_found(adata=oracle_object.adata,
                                                     cluster_column_name=cluster_name_for_GRN_unit,
                                                     embedding_name=oracle_object.embedding_name)

    links.palette = _get_clustercolor_from_anndata(adata=oracle_object.adata,
                                                   cluster_name=cluster_name_for_GRN_unit,
                                                   return_as="palette")

    #links.merge_links()
    links.ALPHA_used = alpha

    links.model_method = model_method

    return links


def _fit_GRN_for_network_analysis(oracle_object, cluster_name_for_GRN_unit, alpha=10, bagging_number=20,
                                  verbose_level=1, test_mode=False, model_method="bagging_ridge", n_jobs=-1):

    # extract information from oracle_object
    gem_imputed = _adata_to_df(oracle_object.adata, "imputed_count")
    gem_imputed_std = standard(gem_imputed)
    cluster_info = oracle_object.adata.obs[cluster_name_for_GRN_unit]
    linkLists = {}

    # setting about verbose
    if verbose_level < 0:
        raise ValueError("varbose_level should be positive number.")
    elif verbose_level == 0:
        loop = np.unique(cluster_info)
        verbose = False
    else:
        loop = tqdm(np.unique(cluster_info))
        if verbose_level <= 1:
            verbose = False
        else:
            verbose = True

    First = True
    for cluster in loop:
        if (not test_mode) | First:
            First = False
            if verbose:
                print(f"Inferring GRN for {cluster}...")

            cells_in_the_cluster_bool = (cluster_info == cluster)
            gem_ = gem_imputed[cells_in_the_cluster_bool]
            gem_std = gem_imputed_std[cells_in_the_cluster_bool]


            tn_ = Net(gene_expression_matrix=gem_,
                         gem_standerdized=gem_std,
                         TFinfo_dic=oracle_object.TFdict,
                         verbose=False)
            tn_.fit_All_genes(bagging_number=bagging_number,
                              model_method=model_method,
                              alpha=alpha,
                              verbose=verbose,
                              n_jobs=n_jobs)


            #oracle_object.linkMat[cluster] = tn_.returnResultAs_TGxTFs("coef_abs")
            tn_.updateLinkList(verbose=False)
            linkLists[cluster] = tn_.linkList.copy()

    return linkLists



# -*- coding: utf-8 -*-
'''

'''


import pandas as pd

SUPPORTED_REF_GENOME = \
    pd.DataFrame([["Human", "hg38", "UCSC"],
                  ["Human", "hg19", "UCSC"],
                  ["Mouse", 'mm39', "UCSC"],
                  ["Mouse", 'mm10', "UCSC"],
                  ["Mouse", 'mm9', "UCSC"],
                  ["S.cerevisiae", "sacCer2", "UCSC"],
                  ["S.cerevisiae", "sacCer3", "UCSC"],
                  ["Zebrafish", "danRer7", "UCSC"],
                  ["Zebrafish", "danRer10", "UCSC"],
                  ["Zebrafish", "danRer11", "UCSC"],
                  ["Xenopus tropicalis", "xenTro2", "UCSC"],
                  ["Xenopus tropicalis", "xenTro3", "UCSC"],
                  ["Xenopus laevis", "Xenopus_laevis_v10.1", "NCBI"],
                  ["Rat", "rn4", "UCSC"],
                  ["Rat", "rn5", "UCSC"],
                  ["Rat", "rn6", "UCSC"],
                  ["Drosophila", "dm3", "UCSC"],
                  ["Drosophila", "dm6", "UCSC"],
                  ["C.elegans", "ce6", "UCSC"],
                  ["C.elegans", "ce10", "UCSC"],
                  ["Arabidopsis", "TAIR10", "Ensembl"],
                  ["Chicken", "galGal4", "UCSC"],
                  ["Chicken", "galGal5", "UCSC"],
                  ["Chicken", "galGal6", "UCSC"],
                  ["Guinea_Pig", "Cavpor3.0", "Ensembl"],
                  ["Pig", "Sscrofa11.1", "Ensembl"],
                  ["Axolotl", "AmexG_v6.0-DD", "Axolotl-omics.org"]
                  ],
                 columns=["species", "ref_genome", "provider"])



import copy
import glob
import time
import os
import shutil
import sys

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns
from tqdm.auto import tqdm

import celloracle as co
from celloracle.applications import Pseudotime_calculator
print("co version: ", co.__version__)

plt.rcParams['figure.figsize'] = [5, 5]
plt.rcParams["savefig.dpi"] = 300

def main():
    test_tutorial_pseudotime()
    
def test_tutorial_pseudotime():
    # Start analysis
    import time
    start = time.time()

    # 1.0 Load data
    print("\n1.0 Load data")
    # Option1. Load oracle data and make pt from oracle object
    oracle = co.data.load_tutorial_oracle_object()
    pt = Pseudotime_calculator(oracle_object=oracle)
    # Option2. Load anndata and make pt from anndata
    adata = co.data.load_Paul2015_data()
    pt = Pseudotime_calculator(adata=adata,
                               obsm_key="X_draw_graph_fa",
                               cluster_column_name="louvain_annot")

    # 2. Pseudotime calculation
    print("\n2. Pseudotime calculation")
    print("Clustering name: ", pt.cluster_column_name)
    print("Cluster list", pt.cluster_list)
    # Check data
    pt.plot_cluster(fontsize=8)
    # Here, clusters can be classified into either MEP lineage or GMP lineage
    clusters_in_ME_lineage = ['Ery_0', 'Ery_1', 'Ery_2', 'Ery_3', 'Ery_4', 'Ery_5',
                              'Ery_6', 'Ery_7', 'Ery_8', 'Ery_9', 'MEP_0', 'Mk_0']
    clusters_in_GM_lineage = ['GMP_0', 'GMP_1', 'GMP_2', 'GMPl_0', 'GMPl_1', 'Gran_0',
                              'Gran_1', 'Gran_2', 'Gran_3', 'Mo_0', 'Mo_1', 'Mo_2']
    # Make a dictionary
    lineage_dictionary = {"Lineage_ME": clusters_in_ME_lineage,
               "Lineage_GM": clusters_in_GM_lineage}
    # Input lineage information into pseudotime object
    pt.set_lineage(lineage_dictionary=lineage_dictionary)
    # Visualize lineage information
    pt.plot_lineages()
    # Set root cell
    # Estimated root cell name for each lineage
    root_cells = {"Lineage_ME": "1539", "Lineage_GM": "2244"}
    pt.set_root_cells(root_cells=root_cells)
    # Check root cell and lineage
    pt.plot_root_cells()
    # Check diffusion map data.
    "X_diffmap" in pt.adata.obsm
    # Calculate pseudotime
    pt.get_pseudotime_per_each_lineage()
    # Check results
    pt.plot_pseudotime(cmap="rainbow")
    # Check result
    pt.adata.obs[["Pseudotime"]].head()

    # 3. Save data
    print("\n3. Save data")

    # If you started calculation with oracle object,
    #  Add calculated pseudotime data to the oracle object
    oracle.adata.obs = pt.adata.obs
    oracle.to_hdf5("oracle_with_pt.celloracle.oracle")
    # If you started calculation with anndata object,
    #  Add calculated pseudotime data to the oracle object
    adata.obs = pt.adata.obs
    adata.write_h5ad("anndata_with_pt.h5ad")

    # Remove files
    print("Remove files")
    for i in ["oracle_with_pt.celloracle.oracle", "anndata_with_pt.h5ad"]:
        os.remove(i)

    elapsed = time.time() - start

    print("Success")
    print(f"Total time: {elapsed/60:.3g} minutes")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from copy import deepcopy
import warnings
import logging
import numpy as np

import matplotlib
import matplotlib.pyplot as plt

import pandas as pd
import os

from sklearn.linear_model import Ridge
from tqdm.auto import tqdm

from ..utility import intersect
#from network_fitting import TransNet as tn



def _do_simulation(coef_matrix, simulation_input, gem, n_propagation):
    """
    Silulate signal propagation in GRNs.

    Args:
        coef_matrix (pandas.DataFrame): 2d matrix that store GRN weights

        simulation_input (pandas.DataFrame): input for simulation

        gem (pandas.DataFrame): input for simulation

        n_propagation (int): number of propagation.

    Returns:
        pandas.DataFrame: simulated gene expression matrix after signal progagation
    """
    delta_input = simulation_input - gem

    delta_simulated = delta_input.copy()
    for i in range(n_propagation):
        delta_simulated = delta_simulated.dot(coef_matrix)
        delta_simulated[delta_input != 0] = delta_input

        # gene expression cannot be negative. adjust delta values to make sure that gene expression are not netavive values.
        gem_tmp = gem + delta_simulated
        gem_tmp[gem_tmp<0] = 0
        delta_simulated = gem_tmp - gem

    gem_simulated = gem + delta_simulated

    return gem_simulated


def _getCoefMatrix(gem, TFdict, alpha=1, verbose=True):
    """
    Calculate GRN and return CoefMatrix (network weights)

    Args:
        gem (pandas.DataFrame): gene expression matrix to calculate GRN

        TFdict (dictionary): python dictionary of potential regulatory gene list

        alpha (float) : strength of regularization in Ridge.

    Returns:
        2d numpy array: numpy array
    """

    genes = gem.columns

    all_genes_in_dict = intersect(gem.columns, list(TFdict.keys()))
    zero_ = pd.Series(np.zeros(len(genes)), index=genes)

    def get_coef(target_gene):
        tmp = zero_.copy()

        # define regGenes
        reggenes = TFdict[target_gene]
        reggenes = intersect(reggenes, genes)

        if target_gene in reggenes:
            reggenes.remove(target_gene)
        if len(reggenes) == 0 :
            tmp[target_gene] = 0
            return(tmp)
        # prepare learning data
        Data = gem[reggenes]
        Label = gem[target_gene]
        # model fitting
        model = Ridge(alpha=alpha, random_state=123)
        model.fit(Data, Label)
        tmp[reggenes] = model.coef_

        return tmp

    li = []
    li_calculated = []
    with tqdm(genes, disable=(verbose==False)) as pbar:
        for i in pbar:
            if not i in all_genes_in_dict:
                tmp = zero_.copy()
                tmp[i] = 0
            else:
                tmp = get_coef(i)
                li_calculated.append(i)
            li.append(tmp)
    coef_matrix = pd.concat(li, axis=1)
    coef_matrix.columns = genes

    if verbose:
        print(f"genes_in_gem: {gem.shape[1]}")
        print(f"models made for {len(li_calculated)} genes")

    return coef_matrix #, li_calculated

def _shuffle_celloracle_GRN_coef_table(coef_dataframe, random_seed=123):
    """
    Shuffle grn coef table. Target genes were shuffled.
    """

    if coef_dataframe.values.max() == 1:
        _correct_coef_table(coef_dataframe=coef_dataframe)
    values = coef_dataframe.values.copy()

    np.random.seed(random_seed)
    random_index = np.arange(values.shape[1])
    np.random.shuffle(random_index)
    random_df = pd.DataFrame(values[:, random_index],
                            index=coef_dataframe.index,
                            columns=coef_dataframe.columns)
    return random_df

def _correct_coef_table(coef_dataframe):
    """
    Delete self regulating edge from coef table.
    """
    for i in range(coef_dataframe.shape[0]):
        coef_dataframe.iloc[i, i] = 0



def _coef_to_active_gene_list(coef_matrix):
    """
    Args:
        coef_matricx (pd.DataFrame): 2d dataframe (gene x gene). This is a result of GRN calculation.

    Return:
        list: list of active gene that have at least one target gene.
    """
    coef = coef_matrix.copy()

    active_TF_list = []
    for i in coef.columns.values:
        coef.loc[i, i] = 0
        if np.any(coef.loc[i, :] != 0):
            active_TF_list.append(i)

    return active_TF_list



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


import seaborn as sns

import os, sys, shutil, importlib, glob
from tqdm.notebook import tqdm

import celloracle as co
from celloracle import motif_analysis as ma
from celloracle.utility import save_as_pickled_object
print("celloracle version: ", co.__version__)

plt.rcParams['figure.figsize'] = (15,7)
plt.rcParams["savefig.dpi"] = 600

def main():
    test_tutorial_motif_scan()
    
def test_tutorial_motif_scan():

    use_small_data = True

    # Start analysis
    import time
    start = time.time()


    # 1. Reference genome data preparation
    print("\n1. Reference genome data preparation")
    # 1.1 Check referenoce genome installation
    ref_genome = "mm10"
    genome_installation = ma.is_genome_installed(ref_genome=ref_genome)
    print(ref_genome, "installation: ", genome_installation)
    ## 1.2. Install reference genome (if refgenome is not installed)
    if not genome_installation:
        import genomepy
        genomepy.install_genome(ref_genome, "UCSC")
    else:
        print(ref_genome, "is installed.")


    # 2. Load data
    # 2.0 Download demo data
    print("\n2. Load data")
    from celloracle.utility.data_download_from_web import download_demo_data
    for i in ["processed_peak_file.csv"]:
        download_demo_data(file=i)

    ## 2.1. Load processed peak data
    # Load annotated peak data.
    peaks = pd.read_csv("processed_peak_file.csv", index_col=0)
    peaks.head()


    ## 2.2. Check data
    def decompose_chrstr(peak_str):
        """
        Args:
            peak_str (str): peak_str. e.g. 'chr1_3094484_3095479'

        Returns:
            tuple: chromosome name, start position, end position
        """

        *chr_, start, end = peak_str.split("_")
        chr_ = "_".join(chr_)
        return chr_, start, end

    from genomepy import Genome

    def check_peak_format(peaks_df, ref_genome):
        """
        Check peak format.
         (1) Check chromosome name.
         (2) Check peak size (length) and remove sort DNA sequences (<5bp)

        """

        df = peaks_df.copy()

        n_peaks_before = df.shape[0]

        # Decompose peaks and make df
        decomposed = [decompose_chrstr(peak_str) for peak_str in df["peak_id"]]
        df_decomposed = pd.DataFrame(np.array(decomposed), index=peaks_df.index)
        df_decomposed.columns = ["chr", "start", "end"]
        df_decomposed["start"] = df_decomposed["start"].astype(int)
        df_decomposed["end"] = df_decomposed["end"].astype(int)

        # Load genome data
        genome_data = Genome(ref_genome)
        all_chr_list = list(genome_data.keys())


        # DNA length check
        lengths = np.abs(df_decomposed["end"] - df_decomposed["start"])


        # Filter peaks with invalid chromosome name
        n_threshold = 5
        df = df[(lengths >= n_threshold) & df_decomposed.chr.isin(all_chr_list)]

        # DNA length check
        lengths = np.abs(df_decomposed["end"] - df_decomposed["start"])

        # Data counting
        n_invalid_length = len(lengths[lengths < n_threshold])
        n_peaks_invalid_chr = n_peaks_before - df_decomposed.chr.isin(all_chr_list).sum()
        n_peaks_after = df.shape[0]


        #
        print("Peaks before filtering: ", n_peaks_before)
        print("Peaks with invalid chr_name: ", n_peaks_invalid_chr)
        print("Peaks with invalid length: ", n_invalid_length)
        print("Peaks after filtering: ", n_peaks_after)

        return df


    peaks = check_peak_format(peaks, ref_genome)


    # 3. Instantiate TFinfo object and search for TF binding motifs
    print("\n3. Instantiate TFinfo object and search for TF binding motifs")
    ## 3.1. Instantiate TFinfo object

    if use_small_data:
        peaks = peaks[:100]
    tfi = ma.TFinfo(peak_data_frame=peaks,
                    ref_genome=ref_genome)
    ## 3.2. Motif scan
    tfi.scan(fpr=0.02,
             motifs=None,  # If you enter None, default motifs will be loaded.
             verbose=True)

    # Save tfinfo object
    tfi.to_hdf5(file_path="test1.celloracle.tfinfo")
    # Check motif scan results
    tfi.scanned_df.head()


    # 4. Filtering motifs
    print("\n4. Filtering motifs")
    # Reset filtering
    tfi.reset_filtering()
    # Do filtering
    tfi.filter_motifs_by_score(threshold=10)
    # Format post-filtering results.
    tfi.make_TFinfo_dataframe_and_dictionary(verbose=True)


    # 5. Get final base GRN
    print("\n5. Get final base GRN")
    df = tfi.to_dataframe()
    df.head()

    # 6. Save result as a dataframe
    print("\n6. Save result as a dataframe")
    df = tfi.to_dataframe()
    df.to_parquet("base_GRN_dataframe.parquet")


    # Remove files
    print("Remove files")
    for i in ["processed_peak_file.csv", "test1.celloracle.tfinfo"]:
        os.remove(i)
        #pass
    elapsed = time.time() - start

    print("Success")
    print(f"Total time: {elapsed/60:.3g} minutes")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
'''
This file contains custom functions for the analysis of ATAC-seq data.
Genomic activity information (peak of ATAC-seq) will be extracted first.
Then the peak DNA sequence will be subjected to TF motif scan.
Finally we will get list of TFs that potentially binds to a specific gene.

Codes were written by Kenji Kamimoto.


'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing

import pandas as pd
import numpy as np
#import matplotlib.pyplot as plt
#import seaborn as sns

import sys
import os
import pickle
import glob
import logging
from copy import deepcopy
from datetime import datetime

from tqdm.auto import tqdm

from pybedtools import BedTool

from ..motif_analysis import __path__ as parent_path
from .process_bed_file import list_peakstr_to_df
from .reference_genomes import SUPPORTED_REF_GENOME

def _load_tss_ref_data(ref_genome):

    """
    Args:
        ref_genome (str): Reference genome name.
            Please contact us through github issue page if you have a request for another referene genome.
    """
    path = os.path.join(parent_path[0], "tss_ref_data", f"{ref_genome}_tss_info.bed")
    return BedTool(fn=path)




def get_tss_info(peak_str_list, ref_genome, verbose=True, custom_tss_file_path=None):
    """
    Get annotation about Transcription Starting Site (TSS).

    Args:
        peak_str_list (list of str): list of peak_id. e.g., [“chr5_0930303_9499409”, “chr11_123445555_123445577”]
        ref_genome (str): reference genome name.
        verbose (bool): verbosity.
        custom_tss_file_path (str): File path to the custom TSS reference bed file. If you just want to use reference genome that are supported in the CellOracle, you don't need to set this parameter.
    """
    if custom_tss_file_path is not None:
        ref = BedTool(fn=custom_tss_file_path)
    else:
        if ref_genome not in SUPPORTED_REF_GENOME.ref_genome.values:
            raise ValueError(f"ref_genome: {ref_genome} is not supported in celloracle. See celloracle.motif_analysis.SUPPORTED_REF_GENOME to get supported ref genome list. If you have a request for a new referencce genome, please post an issue in github issue page.")

        ref = _load_tss_ref_data(ref_genome=ref_genome)

    queue = list_peakstr_to_df(peak_str_list)
    queue = BedTool.from_dataframe(queue)

    annotated = annotate_tss(tss_ref_bed=ref, queue_bed=queue, verbose=verbose)

    return annotated

def annotate_tss(tss_ref_bed, queue_bed, verbose=True):
    """
    Search for peaks that exist in TSS.
    If a peak overlap TSS peak, that peak will be annotated with gene name of the TSS.
    If a peak does not overlap any TSS peak, the peak will be removed and will not be contained in returned bed file.

    Args:
        tss_fer_bed (Bedtool): bedtool object of tss data.
            This bed file should be a bed file with 5 columns: ["chrom", "start", "end", "name", "score", "strand"].
            Gene name of TSS should be stored in "name" column.

        queue_bed (Bedtool): bedtool object.
            This bed file should be a bed file with 3 columns: ["chrom", "start", "end"].

    Returns:
        dataframe: dataframe of peaks.
    """

    # check data structure
    if (tss_ref_bed.to_dataframe().columns != ["chrom", "start", "end", "name", "score", "strand"]).all():
        raise ValueError("tss annotation error")

    if (queue_bed.to_dataframe().columns != ["chrom", "start", "end"]).all():
        raise ValueError("queue bed file error")

    # intersect
    intersected = tss_ref_bed.intersect(queue_bed, wb=True)
    intersected = intersected.to_dataframe()

    # change name
    intersected = intersected.rename(columns={"start": "start_intersected",
                                          "end": "end_intersected",
                                          "name": "gene_short_name"})
    intersected = intersected.rename(columns={"thickStart": "chr", "thickEnd": "start",
                                          "itemRgb": "end"})

    # select data
    intersected = intersected[["chr", "start", "end", "gene_short_name", "strand"]]

    if verbose:
        print(f"que bed peaks: {len(queue_bed.to_dataframe())}")
        print(f"tss peaks in que: {len(intersected)}")
    return intersected



# -*- coding: utf-8 -*-
"""
Data loading function. This is generated by modifying the scanpy's data loading utility function.

"""
from pathlib import Path
import os
import warnings

def download_demo_data(file, git_url=None):
    """
    Download files for the tutorial.
    The downloaded file is saved in the local directory.
    """
    if git_url is None:
        git_url = "https://raw.githubusercontent.com/morris-lab/CellOracle/master/docs/demo_data"
    download_data_if_data_not_exist(path=file,
                                    backup_url=os.path.join(git_url, file))


def download_data_if_data_not_exist(path, backup_url):
    """Check if file is present otherwise download."""

    path = Path(path)
    if not path.is_file():
        _download(url=backup_url, path=path)


def _download(path: Path, url: str):
    print(f"Data not found in the local folder. Loading data from github. Data will be saved at {path.parent}")

    os.makedirs(path.parent, exist_ok=True)

    try:
        import ipywidgets
        from tqdm.auto import tqdm
    except ImportError:
        from tqdm import tqdm

    from urllib.request import urlopen, Request
    from urllib.error import URLError

    blocksize = 1024 * 8
    blocknum = 0

    try:
        req = Request(url)

        try:
            open_url = urlopen(req)
        except URLError:
            warnings.warn('Failed to open the url with default certificates, trying with certifi.')

            from certifi import where
            from ssl import create_default_context
            open_url = urlopen(req, context=create_default_context(cafile=where()))


        with open_url as resp:
            total = resp.info().get("content-length", None)
            with tqdm(
                unit="B",
                unit_scale=True,
                miniters=1,
                unit_divisor=1024,
                total=total if total is None else int(total),
            ) as t, path.open("wb") as f:
                block = resp.read(blocksize)
                while block:
                    f.write(block)
                    blocknum += 1
                    t.update(len(block))
                    block = resp.read(blocksize)

    except (KeyboardInterrupt, Exception):
        # Make sure file doesn’t exist half-downloaded
        if path.is_file():
            path.unlink()
        raise




import os
import pandas as pd
#from warnings import warn

try:
    from importlib.metadata import version
except:
    from importlib_metadata import version

from ..utility import __path__ as PATH


def _load_requirements():
    path = os.path.abspath(PATH[0])
    REQUIREMENTS_PATH = os.path.join(path, "requirements.txt")

    with open(REQUIREMENTS_PATH) as f:
        requirements = f.read().splitlines()

    return requirements


def _get_package_names(requirements):
    names = []
    for i in requirements:
        if ">" in i:
            i = i.split(">")[0]
        elif "=" in i:
            i = i.split("=")[0]
        names.append(i)

    return names

def _get_installed_version(requirements):
    names = _get_package_names(requirements)
    installed_versions = []
    for i in names:
        try:
            v = version(i)
        except:
            v = "not_found"
        installed_versions.append(v)

    return installed_versions

def _get_required_version(requirements):
    versions = []
    for i in requirements:
        if ">=" in i:
            i = i.split(">=")[-1]
        elif ">" in i:
            i = i.split(">")[-1]
        elif "=" in i:
            i = i.split("=")[-1]
        else:
            i = "auto"
        versions.append(i)

    return versions

def _is_version_OK(que, ref):

    try:
        answer = True
        if ref == "auto":
            if que == "not_found":
                answer = False
            else:
                pass
        else:

            que_ = que.split(".")
            ref_ = ref.split(".")


            for q, r in zip(que_, ref_):
                #print(q, r)
                if int(q) < int(r):
                    answer = False
                    break
                elif int(q) > int(r):
                    answer = True
                    break
                else:
                    pass
        return answer
    except:
        return "unknown"

def check_python_requirements(return_detail=True, print_warning=True):
    """
    Check installation status and requirements of dependant libraries.
    """

    try:
        REQUIREMENTS = _load_requirements()
    except:
        if print_warning:
            print("Could not check requirements.")
        return None

    status = pd.DataFrame({"package_name": _get_package_names(REQUIREMENTS),
                           "installed_version": _get_installed_version(REQUIREMENTS),
                            "required_version": _get_required_version(REQUIREMENTS)})
    status["requirement_satisfied"] = \
    [_is_version_OK(ref, que) for ref, que in zip(status["installed_version"],
                                                 status["required_version"])]

    if print_warning:
        n_not_ok = (status["requirement_satisfied"] == False).sum()
        if n_not_ok == 1:
            print(f"{n_not_ok} package does not meet CellOracle requirement.")
        elif n_not_ok >= 2:
            print(f"{n_not_ok} packages do not meet CellOracle requirement.")

        for i, (name, installed, required, isok) in status.iterrows():
            if isok is False:
                print(f" Your {name} version is {installed}. Please install {REQUIREMENTS[i]}")
    if return_detail:
        return status



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from os import path
import re

with open("requirements.txt") as f:
    required = f.read().splitlines()

with open('README.rst') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

root_dir = path.abspath(path.dirname(__file__))
package_name = "celloracle"

with open(path.join(root_dir, package_name, '__init__.py')) as f:
    init_text = f.read()
    license = re.search(r'__license__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    author = re.search(r'__author__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    author_email = re.search(r'__author_email__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    url = re.search(r'__url__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)


with open(path.join(root_dir, package_name, 'version.py')) as f:
    init_text = f.read()
    version = re.search(r'__version__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)

# Start install process
setup(
    name=package_name,
    version=version, #
    description='in silico gene perturbation analysis and GRN analysis with single cell data',
    long_description=readme,
    keywords='scRNA-seq, GRN, simulation, gene perturbation',
    python_requires='>=3.6',
    classifiers=[# How mature is this project? Common values are
                #   3 - Alpha
                #   4 - Beta
                #   5 - Production/Stable
                'Development Status :: 4 - Beta',

                # Indicate who your project is intended for
                'Intended Audience :: Developers',
                'Topic :: Software Development :: Build Tools',

                # Pick your license as you wish (should match "license" above)
                # 'License :: OSI Approved :: MIT License',

                # Specify the Python versions you support here. In particular, ensure
                # that you indicate whether you support Python 2, Python 3 or both.
                'Programming Language :: Python :: 3.6',
                'Programming Language :: Python :: 3.7',
                'Programming Language :: Python :: 3.8'
            ],
    install_requires=required,
    author=author,
    author_email=author_email,
    url=url,
    license=license,
    package_data={"celloracle": ["go_analysis/data/*.txt", "go_analysis/data/*.obo",
                                 "data_conversion/*.R",
                                 "motif_analysis/tss_ref_data/*.bed",
                                 #"data/TFinfo_data/*.txt", "data/TFinfo_data/*.parquet",
                                 #"data/motif_data/*.txt", "data/motif_data/*.pfm",
                                 #"data/anndata/*.h5ad",
                                 #"data/tutorial_data/*.celloracle.oracle", "data/tutorial_data/*.celloracle.links",
                                 #"data/promoter_base_GRN/*.parquet",
                                 "network_analysis/rscripts_for_network_analysis/*.R",
                                 "utility/requirements.txt"]},
    packages=["celloracle", "celloracle.data_conversion", "celloracle.network", "celloracle.trajectory",
              "celloracle.data", "celloracle.go_analysis", "celloracle.oracle_utility",
              "celloracle.motif_analysis", "celloracle.network_analysis", "celloracle.utility", "celloracle.applications", "celloracle.visualizations"],
    entry_points={'console_scripts':['seuratToAnndata = celloracle.data_conversion.process_seurat_object:main']}
)



# -*- coding: utf-8 -*-



from ..network_analysis import load_links
from ..network import load_net
from ..trajectory.oracle_core import load_oracle
from ..motif_analysis.tfinfo_core import load_TFinfo
from ..applications.differentiation_flow import load_gradient

def load_hdf5(file_path, object_class_name=None):
    """
    Load an object of celloracle's custom class that was saved as hdf5.

    Args:
        file_path (str): file_path.
        object_class_name (str): Types of object.
            If it is None, object class will be identified from the extension of file_name.
            Default is None.
    """

    # identify object class type
    if object_class_name is None:
        object_class_name = file_path.split(".")[-1]
    object_class_name = object_class_name.capitalize()

    # load object
    if object_class_name == "Links":
        obj = load_links(file_path=file_path)

    elif object_class_name == "Net":
        obj = load_net(file_path=file_path)

    elif object_class_name == "Oracle":
        obj = load_oracle(file_path=file_path)

    elif object_class_name in ["Tfinfo", "tfinfo"]:
        obj = load_TFinfo(file_path=file_path)

    elif object_class_name == "Gradient":
        obj = load_gradient(file_path=file_path)

    else:
        print(f"object_class_name: {object_class_name} was not in the loading option")
        raise ValueError("File type identification failed. Please enter 'object_class_name' manually.")

    return obj



# -*- coding: utf-8 -*-



import io
import logging
import os
import pickle
import subprocess
import sys

import pandas as pd
import numpy as np
from tqdm.auto import tqdm
import matplotlib.pyplot as plt

from sklearn.preprocessing import StandardScaler

from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.decomposition import PCA
from sklearn.metrics import r2_score

def save_as_pickled_object(obj, filepath):
    """
    Save any object using pickle.

    Args:
        obj (any python object): python object.

        filepath (str): file path.
    """
    max_bytes = 2**31 - 1
    bytes_out = pickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL)
    n_bytes = sys.getsizeof(bytes_out)
    with open(filepath, 'wb') as f_out:
        for idx in range(0, n_bytes, max_bytes):
            f_out.write(bytes_out[idx:idx+max_bytes])


def load_pickled_object(filepath):
    """
    Load pickled object.

    Args:
        filepath (str): file path.

    Returns:
        python object: loaded object.

    """
    max_bytes = 2**31 - 1
    input_size = os.path.getsize(filepath)
    bytes_in = bytearray(0)
    with open(filepath, 'rb') as f_in:
        for _ in range(0, input_size, max_bytes):
            bytes_in += f_in.read(max_bytes)
    obj = pickle.loads(bytes_in)

    return obj



def intersect(list1, list2):
    """
    Intersect two list and get components that exists in both list.

    Args:
        list1 (list): input list.
        list2 (list): input list.

    Returns:
        list: intersected list.

    """
    inter_list = list(set(list1).intersection(list2))
    return(inter_list)


def inverse_dictionary(dictionary, verbose=True, return_value_as_numpy=False):
    """
    Make inverse dictionary.
    See examples below for detail.

    Args:
        dictionary (dict): python dictionary

        verbose (bool): Whether to show progress bar.

        return_value_as_numpy (bool): Whether to convert values into numpy array.

    Returns:
        dict: Python dictionary.

    Examples:
        >>> dic = {"a": [1, 2, 3], "b": [2, 3, 4]}
        >>> inverse_dictionary(dic)
        {1: ['a'], 2: ['a', 'b'], 3: ['a', 'b'], 4: ['b']}

        >>> dic = {"a": [1, 2, 3], "b": [2, 3, 4]}
        >>> inverse_dictionary(dic, return_value_as_numpy=True)
        {1: array(['a'], dtype='<U1'),
         2: array(['a', 'b'], dtype='<U1'),
         3: array(['a', 'b'], dtype='<U1'),
         4: array(['b'], dtype='<U1')}

    """
    def _get_key_list_that_contain_voi_in_its_values(value_of_interest, dictionary):
        keys_list = []
        for i in dictionary:
            values = dictionary[i]
            if value_of_interest in values:
                keys_list.append(i)
        return keys_list

    # get list of all TFs
    all_values = list(dictionary.values())
    all_values = np.unique(np.concatenate(all_values))

    keys_inversed_dict = all_values

    inversed_dictionary = {}

    if verbose:
        loop = tqdm(keys_inversed_dict)
    else:
        loop = keys_inversed_dict

    for key in loop:
        values_in_inversed_dict = _get_key_list_that_contain_voi_in_its_values(key, dictionary)
        values_in_inversed_dict = np.unique(values_in_inversed_dict)
        if not return_value_as_numpy:
            values_in_inversed_dict = list(values_in_inversed_dict)
        inversed_dictionary[key] = values_in_inversed_dict

    return  inversed_dictionary


def exec_process(commands, message=True, wait_finished=True, return_process=True):
    """
    Excute a command. This is a wrapper of "subprocess.Popen"

    Args:
        commands (str): command.
        message (bool): Whether to return a message or not.
        wait_finished (bool): Whether or not to wait for the process to finish. If false, the process will be perfomed in background and the function will finish immediately
        return_process (bool): Whether to return "process".


    """
    my_env = os.environ.copy()
    #print(my_env["PATH"])


    process = subprocess.Popen(commands, stdout=subprocess.PIPE, bufsize=-1,shell=True,
                               env=my_env)

    if message:
        with io.open(process.stdout.fileno(), closefd=False) as stream:
            [print(line.rstrip('\n')) for line in stream]
    if wait_finished:
        # プロセス終了まで待ち、結果を判定する
        process.wait()
        if process.returncode != 0:
            print('Build process aborts.')
            sys.exit(1)
    if return_process:
        return process

def standard(df):
    """
    Standardize value.

    Args:
        df (padas.dataframe): dataframe.

    Returns:
        pandas.dataframe: Data after standardization.

    """
    dt = df.copy()
    #dt = dt.astype("float64")
    # function for data standerdization
    gene_names = dt.columns
    cell_names = dt.index

    scaler = StandardScaler(with_mean=False)
    scaler.fit(dt)
    dt = scaler.transform(dt)

    dt = pd.DataFrame(dt)
    dt.columns = gene_names
    dt.index = cell_names

    return(dt)

## Anndata processing functions

def update_adata(adata):
    # Update Anndata
    # Anndata generated with Scanpy 1.4 or less should be updated with this function
    # This function will be depricated in the future.

    try:
        lo = adata.uns['draw_graph']['params']['layout']
        if isinstance(lo, np.ndarray):
            lo = lo[0]
        adata.uns['draw_graph']['params']['layout'] = lo
    except:
        pass

def adata_to_color_dict(adata, cluster_use):
    """
    Extract color information from adata and returns as dictionary.

    Args:
        adata (anndata): anndata

        cluster_use (str): column name in anndata.obs

    Returns:
        dictionary: python dictionary, key is cluster name, value is clor name
    """
    color_dict = {}
    for i,j in enumerate(adata.obs[cluster_use].cat.categories):
        color_dict[j] = adata.uns[f"{cluster_use}_colors"][i]
    return color_dict



def transfer_all_colors_between_anndata(adata_ref, adata_que):
    """
    Extract all color information from reference anndata and transfer the color into query anndata.

    Args:
        adata_ref (anndata): reference anndata

        adata_que (anndata): query anndata

    """

    # Get all clusters that have colors in adata_ref
    keys = list(adata_ref.uns.keys())
    keys = [i.replace("_colors", "") for i in keys if "_colors" in i]

    keys = [i for i in keys if i in adata_que.obs.columns]

    for cluster_name in keys:
        transfer_color_between_anndata(adata_ref=adata_ref,
                                      adata_que=adata_que,
                                      cluster_name=cluster_name)

    print("Color meta data were transfered for \n")
    print(" ", keys)


def transfer_color_between_anndata(adata_ref, adata_que, cluster_name):
    """
    Extract color information from reference anndata and transfer the color into query anndata.

    Args:
        adata_ref (anndata): reference anndata

        adata_que (anndata): query anndata

        cluster_name (str): cluster name. This information should exist in the anndata.obs.

    """

    # Get color as a dictionary
    dic_ref = adata_to_color_dict(adata_ref, cluster_name)


    # Get color keys from que data
    adata_que.obs[cluster_name] = adata_que.obs[cluster_name].astype("category")
    color_keys = list(adata_que.obs[cluster_name].cat.categories)

    # Convert color list
    colors_after = []
    for i in color_keys:
        color = dic_ref[i]
        colors_after.append(color)

    colors_after = np.array(colors_after)

    # Update color
    adata_que.uns[cluster_name + "_colors"] = colors_after


def knn_data_transferer(adata_ref, adata_que,
                        n_neighbors=20, cluster_name=None, embedding_name=None, adata_true=None,
                        transfer_color=True, n_PCA=30, use_PCA_in_adata=False, meta_data=None):
    """
    Extract categorical information from adata.obs or embedding information from adata.obsm and transfer these information into query anndata.
    In the calculation, KNN is used after PCA.

    Args:
        adata_ref (anndata): reference anndata

        adata_que (anndata): query anndata

        cluster_name (str or list of str): cluster name(s) to be transfered. If you want to transfer multiple data, you can set the cluster names as a list.

        embedding_name (str or list of str): embedding name(s) to be transfered. If you want to transfer multiple data, you can set the embedding names as a list.

        adata_true (str): This argument can be used for the validataion of this algorithm. If you have true answer, you can set it. If you set true answer, the function will return some metrics for benchmarking.

        transfer_color (bool): Whether or not to transfer color data in addition to cluster information.

        n_PCA (int): Number of PCs that will be used for the input of KNN algorithm.

    """
    # 1. Prepare data
    X_train = adata_ref.to_df()
    X_test = adata_que.to_df()

    genes = np.intersect1d(X_train.columns, X_test.columns)
    X_train = X_train[genes]
    X_test = X_test[genes]

    # 2. PCA
    if use_PCA_in_adata:
        X_train_PCA = adata_ref.obsm["X_pca"]
        X_test_PCA = adata_que.obsm["X_pca"]
    else:
        model_PCA = PCA(n_components=n_PCA)
        X_train_PCA = model_PCA.fit_transform(X_train)
        X_test_PCA = model_PCA.transform(X_test)

    # 3. Learning and prediction
    if cluster_name is not None:
        if isinstance(cluster_name, str):
            cluster_name = [cluster_name]
        if isinstance(cluster_name, list):
            for i in  cluster_name:
                model_kncl = KNeighborsClassifier(n_neighbors=n_neighbors)
                model_kncl.fit(X_train_PCA, adata_ref.obs[i])
                pred = model_kncl.predict(X_test_PCA)
                adata_que.obs[i] = pred
                adata_que.obs[i] = adata_que.obs[i].astype("category")

                if transfer_color:
                    transfer_color_between_anndata(adata_ref=adata_ref, adata_que=adata_que, cluster_name=i)

                if adata_true is not None:
                    true = adata_true.obs[i]
                    print(f"accuracy for {i}: ", np.mean(pred == true))
        else:
            raise ValueError("cluster name format error")

    if embedding_name is not None:
        if isinstance(embedding_name, str):
            embedding_name = [embedding_name]
        if isinstance(embedding_name, list):
            for i in  embedding_name:
                model_knreg = KNeighborsRegressor(n_neighbors=n_neighbors)
                model_knreg.fit(X_train_PCA, adata_ref.obsm[i])
                pred = model_knreg.predict(X_test_PCA)
                adata_que.obsm[i] = pred

                if adata_true is not None:
                    true = adata_true.obsm[i]

                    fig, ax = plt.subplots(1, 2)
                    for k in [0, 1]:
                        x = true[:, k]
                        y = pred[:, k]
                        ax[k].scatter(x, y)
                        ax[k].set_xlabel("true")
                        ax[k].set_ylabel("pred")
                        ax[k].set_title(f"r: {r2_score(x, y): .3g}")
                    plt.show()
        else:
            raise ValueError("embedding name format error")

    if meta_data is not None:
        if isinstance(meta_data, str):
            meta_data = [meta_data]
        if isinstance(meta_data, list):
            for i in  meta_data:
                model_knreg = KNeighborsRegressor(n_neighbors=n_neighbors)
                model_knreg.fit(X_train_PCA, adata_ref.obs[i])
                pred = model_knreg.predict(X_test_PCA)
                adata_que.obs[i] = pred

        else:
            raise ValueError("meta_data format error")



# -*- coding: utf-8 -*-
import pandas as pd
from scanpy import AnnData

import celloracle as co

# Mouse scATAC base GRN loading function
def test_load_mouse_scATAC_atlas_base_GRN():
    df = co.data.load_mouse_scATAC_atlas_base_GRN(force_download=False)
    assert isinstance(df, pd.core.frame.DataFrame)

def test_load_mouse_scATAC_atlas_base_GRN_dl():
    df = co.data.load_mouse_scATAC_atlas_base_GRN(force_download=True)
    assert isinstance(df, pd.core.frame.DataFrame)

# Paul dataset anndata loading function
def test_load_Paul2015_data():
    adata = co.data.load_Paul2015_data(force_download=False)
    assert isinstance(adata, AnnData)

def test_load_Paul2015_data_dl():
    adata = co.data.load_Paul2015_data(force_download=True)
    assert isinstance(adata, AnnData)

# Tutorial Links object loading function
def test_load_tutorial_links_object():
    links = co.data.load_tutorial_links_object(force_download=False)
    assert isinstance(links, co.Links)

def test_load_tutorial_links_object_dl():
    links = co.data.load_tutorial_links_object(force_download=True)
    assert isinstance(links, co.Links)

# Tutorial oracle object loading function
def test_load_tutorial_oracle_object():
    oracle = co.data.load_tutorial_oracle_object(force_download=False)
    assert isinstance(oracle, co.Oracle)

def test_load_tutorial_oracle_object_dl():
    oracle = co.data.load_tutorial_oracle_object(force_download=True)
    assert isinstance(oracle, co.Oracle)



# -*- coding: utf-8 -*-
'''
This is a series of custom functions for the inferring of GRN from single cell RNA-seq data.

Codes were written by Kenji Kamimoto.


'''

###########################
### 0. Import libralies ###
###########################


# 0.1. libraries for fundamental data science and data processing

import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from tqdm.auto import tqdm

from .cartography import plot_cartography_kde

#import seaborn as sns

settings = {"save_figure_as": "png"}

def plot_scores_as_rank(links, cluster, n_gene=50, save=None):
    """
    Pick up top n-th genes wich high-network scores and make plots.

    Args:
        links (Links object): See network_analisis.Links class for detail.
        cluster (str): Cluster nome to analyze.
        n_gene (int): Number of genes to plot. Default is 50.
        save (str): Folder path to save plots. If the folde does not exist in the path, the function create the folder.
            If None plots will not be saved. Default is None.
    """
    values = ['degree_centrality_all',
                  'degree_centrality_in', 'degree_centrality_out',
                  'betweenness_centrality',  'eigenvector_centrality']
    for value in values:

        res = links.merged_score[links.merged_score.cluster == cluster]
        res = res[value].sort_values(ascending=False)
        res = res[:n_gene]

        fig = plt.figure()

        plt.scatter(res.values, range(len(res)))
        plt.yticks(range(len(res)), res.index.values)#, rotation=90)
        plt.xlabel(value)
        plt.title(f" {value} \n top {n_gene} in {cluster}")
        plt.gca().invert_yaxis()
        plt.subplots_adjust(left=0.5, right=0.85)

        if not save is None:
            os.makedirs(save, exist_ok=True)
            path = os.path.join(save, f"ranked_values_in_{links.name}_{value}_{links.threshold_number}_in_{cluster}.{settings['save_figure_as']}")
            fig.savefig(path, transparent=True)
        plt.show()


def _plot_goi(x, y, goi, args_annot, scatter=False, x_shift=0.1, y_shift=0.1):
    """
    Plot an annoation to highlight one point in scatter plot.

    Args:
        x (float): Cordinate-x.
        y (float): Cordinate-y.
        args_annot (dictionary): arguments for matplotlib.pyplot.annotate().
        scatter (bool): Whether to plot dot for the point of interest.
        x_shift (float): distance between the annotation and the point of interest in the x-axis.
        y_shift (float): distance between the annotation and the point of interest in the y-axis.
    """

    default = {"size": 10}
    default.update(args_annot)
    args_annot = default.copy()

    arrow_dict = {"width": 0.5, "headwidth": 0.5, "headlength": 1, "color": "black"}
    if scatter:
        plt.scatter(x, y, c="none", edgecolor="black")
    plt.annotate(f"{goi}", xy=(x, y), xytext=(x+x_shift, y+y_shift),
                 color="black", arrowprops=arrow_dict, **args_annot)



def plot_score_comparison_2D(links, value, cluster1, cluster2, percentile=99, annot_shifts=None, save=None, fillna_with_zero=True, plt_show=True):
    """
    Make a scatter plot that shows the relationship of a specific network score in two groups.

    Args:
        links (Links object): See network_analisis.Links class for detail.
        value (srt): The network score to be shown.
        cluster1 (str): Cluster nome to analyze. Network scores in the cluste1 are shown as x-axis.
        cluster2 (str): Cluster nome to analyze. Network scores in the cluste2 are shown as y-axis.
        percentile (float): Genes with a network score above the percentile will be shown with annotation. Default is 99.
        annot_shifts ((float, float)): Shift x and y cordinate for annotations.
        save (str): Folder path to save plots. If the folde does not exist in the path, the function create the folder.
            If None plots will not be saved. Default is None.
    """
    res = links.merged_score[links.merged_score.cluster.isin([cluster1, cluster2])][[value, "cluster"]]
    res = res.reset_index(drop=False)
    piv = pd.pivot_table(res, values=value, columns="cluster", index="index")
    if fillna_with_zero:
        piv = piv.fillna(0)
    else:
        piv = piv.fillna(piv.mean(axis=0))

    goi1 = piv[piv[cluster1] > np.percentile(piv[cluster1].values, percentile)].index
    goi2 = piv[piv[cluster2] > np.percentile(piv[cluster2].values, percentile)].index

    gois = np.union1d(goi1, goi2)

    x, y = piv[cluster1], piv[cluster2]
    plt.scatter(x, y, c="none", edgecolor="black")

    if annot_shifts is None:
        x_shift, y_shift = (x.max() - x.min())*0.03, (y.max() - y.min())*0.03
    else:
        x_shift, y_shift = annot_shifts
    for goi in gois:
        x, y = piv.loc[goi, cluster1], piv.loc[goi, cluster2]
        _plot_goi(x, y, goi, {}, scatter=False, x_shift=x_shift, y_shift=y_shift)

    plt.xlabel(cluster1)
    plt.ylabel(cluster2)
    plt.title(f"{value}")
    if not save is None:
        os.makedirs(save, exist_ok=True)
        path = os.path.join(save, f"values_comparison_in_{links.name}_{value}_{links.threshold_number}_{cluster1}_vs_{cluster2}.{settings['save_figure_as']}")
        plt.savefig(path, transparent=True)
    if plt_show:
        plt.show()




def _test_ver_plot_score_comparison_2D(links, value, cluster1, cluster2, percentile=99,
                             annot_shifts=None, save=None, fillna_with_zero=True,
                             plot_line=True, threshold = 0.25, figsize=(4, 4)):
    from adjustText import adjust_text

    """
    Make a scatter plot that shows the relationship of a specific network score in two groups.

    Args:
        links (Links object): See network_analisis.Links class for detail.
        value (srt): The network score to be shown.
        cluster1 (str): Cluster nome to analyze. Network scores in the cluste1 are shown as x-axis.
        cluster2 (str): Cluster nome to analyze. Network scores in the cluste2 are shown as y-axis.
        percentile (float): Genes with a network score above the percentile will be shown with annotation. Default is 99.
        annot_shifts ((float, float)): Shift x and y cordinate for annotations.
        save (str): Folder path to save plots. If the folde does not exist in the path, the function create the folder.
            If None plots will not be saved. Default is None.
    """
    def rotation_n45(s):
        k = 1/np.sqrt(2)
        x = k*(s[0] + s[1])
        y = k*(-s[0] + s[1])
        return x, y

    def distance_(s):
        x, y = rotation_n45(s)
        return abs(y)

    def rotation_p45(s):
        k = 1/np.sqrt(2)
        x = k*(s[0] - s[1])
        y = k*(s[0] + s[1])
        return x, y

    def distance_1(s):
        x, y = rotation_p45(s)
        return abs(y)

    fig, ax = plt.subplots(figsize=figsize)
    plt.subplots_adjust(left=0.15)
    plt.subplots_adjust(bottom=0.15)


    plt.ticklabel_format(style='sci',axis='y',scilimits=(0,0))
    plt.ticklabel_format(style='sci',axis='x',scilimits=(0,0))

    res = links.merged_score[links.merged_score.cluster.isin([cluster1, cluster2])][[value, "cluster"]]
    res = res.reset_index(drop=False)
    piv = pd.pivot_table(res, values=value, columns="cluster", index="index")
    if fillna_with_zero:
        piv = piv.fillna(0)
    else:
        piv = piv.fillna(piv.mean(axis=0))

    # pick up genes with high score
    goi1 = piv[piv[cluster1] > np.percentile(piv[cluster1].values, percentile)].index
    goi2 = piv[piv[cluster2] > np.percentile(piv[cluster2].values, percentile)].index
    gois = np.union1d(goi1, goi2)


    # plot lines
    x, y = piv[cluster1], piv[cluster2]
    if plot_line:
        # y = x line
        max_ = max(x.max(), y.max())
        plt.plot((0, max_),(0, max_), linestyle=(0, (5, 10)), color="gray", linewidth=0.5)
        # other lines
        plt.plot((0, max_-threshold),(threshold, max_),
                 linestyle="dashed", color="gray", linewidth=0.5, alpha=1)
        plt.plot((threshold, max_), (0, max_-threshold),
                 linestyle="dashed", color="gray", linewidth=0.5, alpha=1)
        plt.plot((0, threshold),(threshold, 0),
                    linestyle="dashed", color="gray", linewidth=0.5, alpha=1)

    # plot all gene
    plt.scatter(x, y, c="lightgray")

    # plot genes that have high scores
    piv["dis_"] = piv.apply(distance_1, axis=1).values
    piv_ = piv[piv.dis_ >= threshold/np.sqrt(2)]
    x, y = piv_[cluster1], piv_[cluster2]
    plt.scatter(x, y, c="dimgray")

    # plot genes that are differentially connected
    piv_["dis"] = piv_.apply(distance_, axis=1).values
    x, y = piv_[piv_.dis>=threshold/np.sqrt(2)][cluster1], piv_[piv_.dis>=threshold/np.sqrt(2)][cluster2]
    gois = np.concatenate([gois, x.index])
    plt.scatter(x, y, c="indianred")

    # plot annotations
    texts = []
    gois = np.unique(gois)
    for goi in gois:
        x, y = piv.loc[goi, cluster1], piv.loc[goi, cluster2]
        plt_text = ax.annotate(goi, (x, y), fontsize=8, color='black')
        texts.append(plt_text)

    adjust_text(texts, arrowprops=dict(arrowstyle='-', color='dimgray', lw=0.5))

    plt.xlabel(cluster1)
    plt.ylabel(cluster2)
    plt.title(f"{value}")
    if not save is None:
        os.makedirs(save, exist_ok=True)
        path = os.path.join(save, f"values_comparison_in_{links.name}_{value}_{links.threshold_number}_{cluster1}_vs_{cluster2}.{settings['save_figure_as']}")
        plt.savefig(path, transparent=True)
    plt.show()

try:
    import plotly.express as px
except:
    pass

def plot_score_comparison_2D_with_plotly(links, value, cluster1, cluster2, fillna_with_zero=True):
    """
    Make a scatter plot that shows the relationship of a specific network score in two groups.

    Args:
        links (Links object): See network_analisis.Links class for detail.
        value (srt): The network score to be shown.
        cluster1 (str): Cluster nome to analyze. Network scores in the cluste1 are shown as x-axis.
        cluster2 (str): Cluster nome to analyze. Network scores in the cluste2 are shown as y-axis.

    """

    try:
        import plotly.express as px
        res = links.merged_score[links.merged_score.cluster.isin([cluster1, cluster2])][[value, "cluster"]]
        res = res.reset_index(drop=False)
        piv = pd.pivot_table(res, values=value, columns="cluster", index="index")
        piv = piv.reset_index(drop=False)

        if fillna_with_zero:
            piv = piv.fillna(0)
        else:
            piv = piv.fillna(piv.mean(axis=0))


        x, y = piv[cluster1], piv[cluster2]

        fig = px.scatter(piv, x=cluster1, y=cluster2,
                         hover_data=['index'], template="plotly_white")

        return fig
    except:
        print("Interactive mode requires plotly. Please install plotly before use.: pip install plotly")




######################
### score dynamics ###
######################
def plot_score_per_cluster(links, goi, save=None, plt_show=True):
    """
    Plot network score for a specific gene.
    This function can be used to compare network score of a specific gene between clusters
    and get insight about the dynamics of the gene.

    Args:
        links (Links object): See network_analisis.Links class for detail.
        goi (srt): Gene name.
        save (str): Folder path to save plots. If the folde does not exist in the path, the function create the folder.
            If None plots will not be saved. Default is None.
    """
    print(goi)
    res = links.merged_score[links.merged_score.index==goi]
    res = res.rename(
        columns={"degree_centrality_all": "degree\ncentrality",
                 "betweenness_centrality": "betweenness\ncentrality",
                 "eigenvector_centrality": "eigenvector\ncentrality"})
    # make plots
    values = [ "degree\ncentrality",  "betweenness\ncentrality",
              "eigenvector\ncentrality"]
    for i, value in zip([1, 2, 3], values):
        plt.subplot(1, 3,  i)
        ax = sns.stripplot(data=res, y="cluster", x=value,
                      size=10, orient="h",linewidth=1, edgecolor="w",
                      order=links.palette.index.values,
                      palette=dict(links.palette.palette))
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.yaxis.grid(True)
        ax.tick_params(bottom=False,
                        left=False,
                        right=False,
                        top=False)
        if i > 1:
            plt.ylabel(None)
            ax.tick_params(labelleft=False)

    if not save is None:
        os.makedirs(save, exist_ok=True)
        path = os.path.join(save,
                           f"score_dynamics_in_{links.name}_{links.threshold_number}_{goi}.{settings['save_figure_as']}")
        plt.savefig(path, transparent=True)
    if plt_show:
        plt.show()


###################
### cartography ###
###################

def plot_cartography_scatter_per_cluster(links, gois=None, clusters=None,
                                         scatter=False, kde=True,
                                         auto_gene_annot=False, percentile=98,
                                         args_dot={}, args_line={},
                                         args_annot={}, save=None):
    """
    Plot gene network cartography in a scatter or kde plot.
    Please read the original paper of gene network cartography for detail.
    https://www.nature.com/articles/nature03288

    Args:
        links (Links object): See network_analisis.Links class for detail.
        gois (list of srt): List of Gene name to highlight.
        clusters (list of str): List of cluster name to analyze. If None, all clusters in Links object will be analyzed.
        scatter (bool): Whether to make scatter plot.
        auto_gene_annot (bool): Whether to pick up genes to make annotation.
        percentile (float): Genes with a network score above the percentile will be shown with annotation. Default is 98.
        args_dot (dictionary): Arguments for scatter plot.
        args_line (dictionary): Arguments for lines in cartography plot.
        args_annot (dictionary): Arguments for annoation in plots.
        save (str): Folder path to save plots. If the folde does not exist in the path, the function create the folder.
            If None plots will not be saved. Default is None.
    """
    if clusters is None:
        clusters = links.cluster

    for cluster in clusters:
        print(cluster)
        data = links.merged_score[links.merged_score.cluster == cluster]
        data = data[["connectivity", "participation"]]
        #data = data[["within_module_degree", "participation_coefficient"]]

        if auto_gene_annot:
            goi1 = data[data["connectivity"] > np.percentile(data["connectivity"].values, percentile)].index
            goi2 = data[data["participation"] > np.percentile(data["participation"].values, percentile)].index

            if gois is None:
                gois_ = np.union1d(goi1, goi2)
            else:
                gois_ = np.union1d(gois, goi1)
                gois_ = np.union1d(gois_, goi2)
        else:
            gois_ = gois

        fig = plt.figure()

        plot_cartography_kde(data, gois_, scatter, kde,
                             args_dot, args_line, args_annot)
        plt.title(f"cartography in {cluster}")
        plt.subplots_adjust(left=0.2, bottom=0.25)

        if not save is None:
            os.makedirs(save, exist_ok=True)
            path = os.path.join(save, f"cartography_in_{links.name}_{links.threshold_number}_{cluster}.{settings['save_figure_as']}")
            fig.savefig(path, transparent=True)
        plt.show()


def plot_cartography_term(links, goi, save=None, plt_show=True):
    """
    Plot the summary of gene network cartography like a heatmap.
    Please read the original paper of gene network cartography for detail.
    https://www.nature.com/articles/nature03288

    Args:
        links (Links object): See network_analisis.Links class for detail.
        gois (list of srt): List of Gene name to highlight.
        save (str): Folder path to save plots. If the folde does not exist in the path, the function create the folder.
            If None plots will not be saved. Default is None.
    """
    print(goi)
    tt = pd.get_dummies(links.merged_score[["cluster", "role"]],columns=["role"])
    tt = tt.loc[goi].set_index("cluster")
    tt.columns = [i.replace("role_", "") for i in tt.columns]

    order = ["Ultra peripheral", "Peripheral", "Connector","Kinless","Provincical Hub","Connector Hub", "Kinless Hub"]

    tt = tt.reindex(index=links.palette.index.values, columns=order).fillna(0)

    sns.heatmap(data=tt, cmap="Blues", cbar=False)
    if not save is None:
        os.makedirs(save, exist_ok=True)
        path = os.path.join(save,
                           f"cartography_role_in_{links.name}_{links.threshold_number}_{goi}.{settings['save_figure_as']}")
        plt.savefig(path, transparent=True)

    if plt_show:
        plt.show()
